METADATA_ENTRY_FRAGMENT = """
fragment MetadataEntryFragment on EventMetadataEntry {
  __typename
  label
  description
  ... on EventPathMetadataEntry {
    path
  }
  ... on EventJsonMetadataEntry {
    jsonString
  }
  ... on EventUrlMetadataEntry {
    url
  }
  ... on EventTextMetadataEntry {
    text
  }
  ... on EventMarkdownMetadataEntry {
    mdStr
  }
  ... on EventPythonArtifactMetadataEntry {
    module
    name
  }
  ... on EventFloatMetadataEntry {
    floatValue
  }
  ... on EventIntMetadataEntry {
    intValue
  }
}
"""

PYTHON_ERROR_FRAGMENT = """
fragment PythonErrorFragment on PythonError {
  __typename
  message
  className
  stack
  cause {
    message
    stack
  }
}
"""

SERIALIZABLE_ERROR_INFO_FRAGMENT = """

"""

EVENT_LOG_ENTRY_FRAGMENT = """
fragment EventLogEntryFragment on EventLogEntry {
    errorInfo
    level
    userMessage
    runId
    timestamp
    stepKey
    pipelineName
    dagsterEvent
}
"""

EVENT_RECORD_FRAGMENT = (
    EVENT_LOG_ENTRY_FRAGMENT
    + """
    fragment EventLogRecordFragment on EventLogRecord {
        storageId
        eventLogEntry {
            ...EventLogEntryFragment
        }
    }
    """
)

CACHED_STATUS_DATA_FRAGMENT = """
fragment CachedStatusDataFragment on AssetStatusCacheValue {
    latestStorageId
    partitionsDefId
    serializedMaterializedPartitionSubset
    serializedFailedPartitionSubset
    serializedInProgressPartitionSubset
    earliestInProgressMaterializationEventId
}
"""

ASSET_ENTRY_FRAGMENT = (
    EVENT_RECORD_FRAGMENT
    + CACHED_STATUS_DATA_FRAGMENT
    + """
    fragment AssetEntryFragment on AssetEntry {
        assetKey {
            path
        }
        lastMaterializationRecord {
            ...EventLogRecordFragment
        }
        lastRunId
        assetDetails {
            lastWipeTimestamp
        }
        cachedStatus {
            ...CachedStatusDataFragment
        }
    }
    """
)

ASSET_RECORD_FRAGMENT = (
    ASSET_ENTRY_FRAGMENT
    + """
    fragment AssetRecordFragment on AssetRecord {
        storageId
        assetEntry {
            ...AssetEntryFragment
        }
    }
    """
)

GET_RECORDS_FOR_RUN_QUERY = (
    EVENT_RECORD_FRAGMENT
    + """
    query getRecordsForRun($runId: String!, $cursor: String, $ofType: String, $ofTypes: [String!], $limit: Int) {
        eventLogs {
            getRecordsForRun(runId: $runId, cursor: $cursor, ofType: $ofType, ofTypes: $ofTypes, limit: $limit) {
                records {
                    ...EventLogRecordFragment
                }
                cursor
                hasMore
            }
        }
    }
    """
)

GET_STATS_FOR_RUN_QUERY = """
    query getStatsForRun($runId: String!) {
        eventLogs {
            getStatsForRun(runId: $runId) {
                runId
                stepsSucceeded
                stepsFailed
                materializations
                expectations
                enqueuedTime
                launchTime
                startTime
                endTime
            }
        }
    }
    """

GET_STEP_STATS_FOR_RUN_QUERY = """
    query getStepStatsForRun($runId: String!, $stepKeys: [String!]) {
        eventLogs {
            getStepStatsForRun(runId: $runId, stepKeys: $stepKeys) {
                runId
                stepKey
                status
                startTime
                endTime
                materializationEvents
                expectationResults
                attempts
                attemptsList
                markers
            }
        }
    }
    """

IS_PERSISTENT_QUERY = """
    query isPersistent {
        eventLogs {
            isPersistent
        }
    }
    """

IS_ASSET_AWARE_QUERY = """
    query isAssetAware {
        eventLogs {
            isAssetAware
        }
    }
    """

GET_EVENT_RECORDS_QUERY = (
    EVENT_RECORD_FRAGMENT
    + """
    query getEventRecords($eventRecordsFilter: EventRecordsFilter, $limit: Int, $ascending: Boolean) {
        eventLogs {
            getEventRecords(
                eventRecordsFilter: $eventRecordsFilter,
                limit: $limit,
                ascending: $ascending,
            ) {
                ...EventLogRecordFragment
            }
        }
    }
    """
)
HAS_ASSET_KEY_QUERY = """
    query hasAssetKey($assetKey: String!) {
        eventLogs {
            hasAssetKey(assetKey: $assetKey)
        }
    }
    """

GET_LATEST_MATERIALIZATION_EVENTS_QUERY = (
    EVENT_LOG_ENTRY_FRAGMENT
    + """
    query getLatestMaterializationEvents($assetKeys: [String!]!) {
        eventLogs {
            getLatestMaterializationEvents(assetKeys: $assetKeys) {
                ...EventLogEntryFragment
            }
        }
    }
    """
)

GET_ASSET_RECORDS_QUERY = (
    ASSET_RECORD_FRAGMENT
    + """
    query getAssetRecords($assetKeys: [String!]) {
        eventLogs {
            getAssetRecords(assetKeys: $assetKeys) {
                ...AssetRecordFragment
            }
        }
    }
    """
)

GET_ALL_ASSET_KEYS_QUERY = """
    query getAllAssetKeys {
        eventLogs {
            getAllAssetKeys
        }
    }
    """

UPDATE_ASSET_CACHED_STATUS_DATA_MUTATION = """
    mutation updateAssetCachedStatusData($assetKey: String!, $cacheValues: AssetStatusCacheValueInput!) {
        eventLogs {
            UpdateAssetCachedStatusData(assetKey: $assetKey, cacheValues: $cacheValues) {
                ok
            }
        }
    }
"""

GET_MATERIALIZED_PARTITIONS = """
    query getMaterializedPartitions($assetKey: String!, $beforeCursor: BigInt, $afterCursor: BigInt) {
        eventLogs {
            getMaterializedPartitions(assetKey: $assetKey, beforeCursor: $beforeCursor, afterCursor: $afterCursor)
        }
    }
"""

GET_MATERIALIZATION_COUNT_BY_PARTITION = """
    query getMaterializationCountByPartition($assetKeys: [String!]!, $afterCursor: BigInt) {
        eventLogs {
            getMaterializationCountByPartition(assetKeys: $assetKeys, afterCursor: $afterCursor) {
                ... on AssetMaterializationCountByPartition {
                    assetKey
                    materializationCountByPartition {
                        partition
                        materializationCount
                    }
                }
            }
        }
    }
    """

GET_LATEST_STORAGE_ID_BY_PARTITION = """
    query getLatestStorageIdByPartition($assetKey: String!, $eventType: String!) {
        eventLogs {
            getLatestStorageIdByPartition(assetKey: $assetKey, eventType: $eventType) {
                ... on StorageIdByPartition {
                    partition
                    storageId
                }
            }
        }
    }
"""

GET_LATEST_TAGS_BY_PARTITION = """
    query getLatestTagsByPartition(
        $assetKey: String!,
        $eventType: String!,
        $tagKeys: [String!],
        $assetPartitions: [String!],
        $beforeCursor: BigInt,
        $afterCursor: BigInt,
    ) {
        eventLogs {
            getLatestTagsByPartition(
                assetKey: $assetKey,
                eventType: $eventType,
                tagKeys: $tagKeys,
                assetPartitions: $assetPartitions,
                beforeCursor: $beforeCursor,
                afterCursor: $afterCursor,
            ) {
                ... on TagByPartition {
                    partition
                    key
                    value
                }
            }
        }
    }
"""

GET_LATEST_ASSET_PARTITION_MATERIALIZATION_ATTEMPTS_WITHOUT_MATERIALIZATIONS = """
    query getLatestAssetPartitionMaterializationAttemptsWithoutMaterializations($assetKey: String!, $afterStorageId: BigInt) {
        eventLogs {
            getLatestAssetPartitionMaterializationAttemptsWithoutMaterializations(assetKey: $assetKey, afterStorageId: $afterStorageId)
        }
    }
    """

GET_EVENT_TAGS_FOR_ASSET = """
query getAssetEventTags($assetKey: String!, $filterTags: [AssetFilterTagInput!], $filterEventId: BigInt) {
    eventLogs {
        getAssetEventTags(assetKey: $assetKey, filterTags: $filterTags, filterEventId: $filterEventId) {
            ... on AssetEventTags {
                tags {
                    key
                    value
                }
            }
        }
    }
}
"""


STORE_EVENT_MUTATION = """
    mutation StoreEvent($eventRecord: EventLogEntryInput!) {
        eventLogs {
            StoreEvent(eventRecord: $eventRecord) {
                ok
            }
        }
    }
    """

DELETE_EVENTS_MUTATION = """
    mutation DeleteEvents($runId: String!) {
        eventLogs {
            DeleteEvents(runId: $runId) {
                ok
            }
        }
    }
    """

UPGRADE_EVENT_LOG_STORAGE_MUTATION = """
    mutation UpgradeEventLogStorage {
        eventLogs {
            Upgrade {
                ok
            }
        }
    }
    """

REINDEX_MUTATION = """
    mutation Reindex {
        eventLogs {
            Reindex {
                ok
            }
        }
    }
"""

WIPE_EVENT_LOG_STORAGE_MUTATION = """
    mutation WipeEventLogStorage {
        eventLogs {
            Wipe {
                ok
            }
        }
    }
"""

ENABLE_SECONDARY_INDEX_MUTATION = """
    mutation EnableSecondaryIndex($name: String!, $runId: String) {
        eventLogs {
            EnableSecondaryIndex(name: $name, runId: $runId) {
                ok
            }
        }
    }
"""

WIPE_ASSET_MUTATION = """
    mutation WipeAsset($assetKey: String!) {
        eventLogs {
            WipeAsset(assetKey: $assetKey) {
                ok
            }
        }
    }
"""

WIPE_ASSET_CACHED_STATUS_DATA_MUTATION = """
    mutation WipeAssetCachedStatusData($assetKey: String!) {
        eventLogs {
            WipeAssetCachedStatusData(assetKey: $assetKey) {
                ok
            }
        }
    }
"""

GET_DYNAMIC_PARTITIONS_QUERY = """
query getDynamicPartitions($partitionsDefName: String!) {
    eventLogs {
        getDynamicPartitions(partitionsDefName: $partitionsDefName)
    }
}
"""

HAS_DYNAMIC_PARTITION_QUERY = """
query hasDynamicPartition($partitionsDefName: String!, $partitionKey: String!) {
    eventLogs {
        hasDynamicPartition(partitionsDefName: $partitionsDefName, partitionKey: $partitionKey)
    }
}
"""

ADD_DYNAMIC_PARTITIONS_MUTATION = """
    mutation AddDynamicPartitions($partitionsDefName: String!, $partitionKeys: [String!]!) {
        eventLogs {
            AddDynamicPartitions(partitionsDefName: $partitionsDefName, partitionKeys: $partitionKeys) {
                ok
            }
        }
    }
"""

DELETE_DYNAMIC_PARTITION_MUTATION = """
    mutation DeleteDynamicPartition($partitionsDefName: String!, $partitionKey: String!) {
        eventLogs {
            DeleteDynamicPartition(partitionsDefName: $partitionsDefName, partitionKey: $partitionKey) {
                ok
            }
        }
    }
"""

SET_CONCURRENCY_SLOTS_MUTATION = (
    PYTHON_ERROR_FRAGMENT
    + """
mutation SetConcurrencySlots($concurrencyKey: String!, $num: Int!) {
    eventLogs {
        SetConcurrencySlots(concurrencyKey: $concurrencyKey, num: $num) {
            success
            error {
                ...PythonErrorFragment
            }
        }
    }
}
"""
)

DELETE_CONCURRENCY_LIMIT_MUTATION = (
    PYTHON_ERROR_FRAGMENT
    + """
mutation DeleteConcurrencyLimit($concurrencyKey: String!) {
    eventLogs {
        DeleteConcurrencyLimit(concurrencyKey: $concurrencyKey) {
            success
            error {
                ...PythonErrorFragment
            }
        }
    }
}
"""
)
INITIALIZE_CONCURRENCY_LIMIT_MUTATION = (
    PYTHON_ERROR_FRAGMENT
    + """
mutation InitializeConcurrencyLimit($concurrencyKey: String!) {
    eventLogs {
        InitializeConcurrencyLimit(concurrencyKey: $concurrencyKey) {
            success
            error {
                ...PythonErrorFragment
            }
        }
    }
}
"""
)

GET_CONCURRENCY_KEYS_QUERY = """
query getConcurrencyKeys {
    eventLogs {
        getConcurrencyKeys
    }
}
"""

GET_CONCURRENCY_INFO_QUERY = """
query getConcurrencyInfo($concurrencyKey: String!) {
    eventLogs {
        getConcurrencyInfo(concurrencyKey: $concurrencyKey) {
            concurrencyKey
            slotCount
            claimedSlots {
                runId
                stepKey
            }
            pendingSteps {
                runId
                stepKey
                enqueuedTimestamp
                assignedTimestamp
                priority
            }
        }
    }
}
"""

CLAIM_CONCURRENCY_SLOT_MUTATION = """
mutation ClaimConcurrencySlot($concurrencyKey: String!, $runId: String!, $stepKey: String!, $priority: Int!) {
    eventLogs {
        ClaimConcurrencySlot(concurrencyKey: $concurrencyKey, runId: $runId, stepKey: $stepKey, priority: $priority) {
            status {
                slotStatus
                priority
                assignedTimestamp
                enqueuedTimestamp
            }
        }
    }
}
"""

CHECK_CONCURRENCY_CLAIM_QUERY = """
query getCheckConcurrencyClaim($concurrencyKey: String!, $runId: String!, $stepKey: String!) {
    eventLogs {
        getCheckConcurrencyClaim(concurrencyKey: $concurrencyKey, runId: $runId, stepKey: $stepKey) {
            slotStatus
            priority
            assignedTimestamp
            enqueuedTimestamp
        }
    }
}
"""

FREE_CONCURRENCY_SLOTS_FOR_RUN_MUTATION = """
mutation FreeConcurrencySlotsForRun($runId: String!) {
    eventLogs {
        FreeConcurrencySlotsForRun(runId: $runId) {
            success
        }
    }
}
"""

FREE_CONCURRENCY_SLOT_FOR_STEP_MUTATION = """
mutation FreeConcurrencySlotForStep($runId: String!, $stepKey: String!) {
    eventLogs {
        FreeConcurrencySlotForStep(runId: $runId, stepKey: $stepKey) {
            success
        }
    }
}
"""


FETCH_MATERIALIZATIONS_QUERY = (
    EVENT_RECORD_FRAGMENT
    + """
query fetchMaterializationsQuery($recordsFilter: AssetRecordsFilter!, $limit: Int, $cursor: String, $ascending: Boolean) {
    eventLogs {
        fetchMaterializations(recordsFilter: $recordsFilter, limit: $limit, cursor: $cursor, ascending: $ascending) {
            records {
                ...EventLogRecordFragment
            }
            cursor
            hasMore
        }
    }
}
"""
)

FETCH_OBSERVATIONS_QUERY = (
    EVENT_RECORD_FRAGMENT
    + """
query fetchObservationsQuery($recordsFilter: AssetRecordsFilter!, $limit: Int, $cursor: String, $ascending: Boolean) {
    eventLogs {
        fetchObservations(recordsFilter: $recordsFilter, limit: $limit, cursor: $cursor, ascending: $ascending) {
            records {
                ...EventLogRecordFragment
            }
            cursor
            hasMore
        }
    }
}
"""
)

FETCH_PLANNED_MATERIALIZATIONS = (
    EVENT_RECORD_FRAGMENT
    + """
query fetchPlannedMaterializationsQuery($recordsFilter: AssetRecordsFilter!, $limit: Int, $cursor: String, $ascending: Boolean) {
    eventLogs {
        fetchPlannedMaterializations(recordsFilter: $recordsFilter, limit: $limit, cursor: $cursor, ascending: $ascending) {
            records {
                ...EventLogRecordFragment
            }
            cursor
            hasMore
        }
    }
}
"""
)

GET_RUN_STATUS_CHANGE_EVENTS_QUERY = (
    EVENT_RECORD_FRAGMENT
    + """
query getRunStatusChangeEventRecordsQuery($recordsFilter: RunStatusChangeRecordsFilter!, $limit: Int, $cursor: String, $ascending: Boolean) {
    eventLogs {
        getRunStatusChangeEventRecords(recordsFilter: $recordsFilter, limit: $limit, cursor: $cursor, ascending: $ascending) {
            records {
                ...EventLogRecordFragment
            }
            cursor
            hasMore
        }
    }
}
"""
)

GET_LATEST_PLANNED_MATERIALIZATION_INFO = """
    query getLatestPlannedMaterializationInfo($assetKey: String!, $partition: String) {
        eventLogs {
            getLatestPlannedMaterializationInfo(assetKey: $assetKey, partition: $partition) {
                storageId
                runId
            }
        }
    }

"""



PUT_CLOUD_METRICS_MUTATION = """
mutation CreateOrUpdateExtenalMetrics(
  $metrics: [MetricInputs]!) {
  createOrUpdateMetrics(metrics: $metrics) {
    __typename
    ... on CreateOrUpdateMetricsSuccess {
      status
      __typename
    }
    ...MetricsFailedFragment
    ...UnauthorizedErrorFragment
    ...PythonErrorFragment
  }
}
fragment PythonErrorFragment on PythonError {
  __typename
  message
  stack
  causes {
    message
    stack
    __typename
  }
}
fragment MetricsFailedFragment on CreateOrUpdateMetricsFailed {
  __typename
  message
}
fragment UnauthorizedErrorFragment on UnauthorizedError {
  __typename
  message
}
"""


PUT_COST_INFORMATION_MUTATION = """
mutation SubmitCostInformation(
  $metricName: String!, $costInfo: [CostInformation!], $start: Float!, $end: Float!) {
  submitCostInformationForMetrics(metricName: $metricName, costInfo: $costInfo, start: $start, end: $end) {
    __typename
    ... on CreateOrUpdateMetricsSuccess {
      status
      __typename
    }
    ...MetricsFailedFragment
    ...UnauthorizedErrorFragment
    ...PythonErrorFragment
  }
}
fragment PythonErrorFragment on PythonError {
  __typename
  message
  stack
  causes {
    message
    stack
    __typename
  }
}
fragment MetricsFailedFragment on CreateOrUpdateMetricsFailed {
  __typename
  message
}
fragment UnauthorizedErrorFragment on UnauthorizedError {
  __typename
  message
}
"""



import os
import tempfile
from typing import Any, Dict, Generator, List, NamedTuple, Optional, Tuple, Union

import dagster._check as check
import requests
from dagster import AssetExecutionContext, DagsterInstance, OpExecutionContext
from dagster._annotations import experimental
from dagster_cloud_cli.core.errors import raise_http_error
from dagster_cloud_cli.core.headers.auth import DagsterCloudInstanceScope
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

from dagster_cloud.instance import DagsterCloudAgentInstance

from .query import PUT_CLOUD_METRICS_MUTATION, PUT_COST_INFORMATION_MUTATION


def _chunks(chunk_list: List[Any], length: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(chunk_list), length):
        yield chunk_list[i : i + length]


@experimental
class DagsterMetric(NamedTuple):
    """Experimental: This class gives information about a Metric.

    Args:
        metric_name (str): name of the metric
        metric_value (float): value of the metric
    """

    metric_name: str
    metric_value: float


def query_graphql_from_instance(
    instance: DagsterInstance, query_text: str, variables=None
) -> Dict[str, Any]:
    headers = {}

    url, cloud_token = get_url_and_token_from_instance(instance)

    headers["Dagster-Cloud-API-Token"] = cloud_token

    transport = RequestsHTTPTransport(
        url=url,
        use_json=True,
        timeout=300,
        headers={"Dagster-Cloud-Api-Token": cloud_token},
    )
    client = Client(transport=transport, fetch_schema_from_transport=True)
    return client.execute(gql(query_text), variable_values=variables or dict())


def get_url_and_token_from_instance(instance: DagsterInstance) -> Tuple[str, str]:
    if not isinstance(instance, DagsterCloudAgentInstance):
        raise RuntimeError("This asset only functions in a running Dagster Cloud instance")

    return f"{instance.dagit_url}graphql", instance.dagster_cloud_agent_token


def _cost_information_chunk_size() -> int:
    return 10000


def chunk_by_opaque_id(
    chunk_list: List[Tuple[str, float]], length: int
) -> Generator[List[Tuple[str, float]], None, None]:
    """Yield chunks of data. Groups by opaque_id so that
    all metrics for a given opaque_id are in the same chunk.
    Chunks will be of size `length` or less unless a single
    opaque_id has more than `length` metrics.
    """
    sorted_chunk_list = sorted(chunk_list, key=lambda x: x[0]) + [("marker", 0.0)]

    current_opaque_id = None
    last_opaque_id_boundary = 0
    last_chunk_boundary = 0
    for i, (opaque_id, _) in enumerate(sorted_chunk_list):
        # If we've seen enough metrics to fill a chunk, yield the chunk minus the
        # current opaque_id, and start a new chunk
        if i - last_chunk_boundary > length:
            yield sorted_chunk_list[last_chunk_boundary:last_opaque_id_boundary]
            last_chunk_boundary = last_opaque_id_boundary

        if opaque_id != current_opaque_id:
            current_opaque_id = opaque_id
            last_opaque_id_boundary = i

    sorted_chunk_list = sorted_chunk_list[:-1]
    if last_chunk_boundary < len(sorted_chunk_list):
        yield sorted_chunk_list[last_chunk_boundary:]


def get_post_request_params(
    instance: DagsterInstance,
) -> Tuple[requests.Session, str, Dict[str, str], int, Optional[Dict[str, str]]]:
    if not isinstance(instance, DagsterCloudAgentInstance):
        raise RuntimeError("This asset only functions in a running Dagster Cloud instance")

    return (
        instance.rest_requests_session,
        instance.dagster_cloud_gen_insights_url_url,
        instance.dagster_cloud_api_headers(DagsterCloudInstanceScope.DEPLOYMENT),
        instance.dagster_cloud_api_timeout,
        instance.dagster_cloud_api_proxies,
    )


def upload_cost_information(
    context: Union[OpExecutionContext, AssetExecutionContext],
    metric_name: str,
    cost_information: List[Tuple[str, float, str]],
):
    import pyarrow as pa
    import pyarrow.parquet as pq

    with tempfile.TemporaryDirectory() as temp_dir:
        opaque_ids = pa.array([opaque_id for opaque_id, _, _ in cost_information])
        costs = pa.array([cost for _, cost, _ in cost_information])
        query_ids = pa.array([query_id for _, _, query_id in cost_information])
        metric_names = pa.array([metric_name for _, _, _ in cost_information])

        cost_pq_file = os.path.join(temp_dir, "cost.parquet")
        pq.write_table(
            pa.Table.from_arrays(
                [opaque_ids, costs, metric_names, query_ids],
                ["opaque_id", "cost", "metric_name", "query_id"],
            ),
            cost_pq_file,
        )

        instance = context.instance
        session, url, headers, timeout, proxies = get_post_request_params(instance)

        resp = session.post(url, headers=headers, timeout=timeout, proxies=proxies)
        raise_http_error(resp)
        resp_data = resp.json()

        assert "url" in resp_data and "fields" in resp_data, resp_data

        with open(cost_pq_file, "rb") as f:
            session.post(
                resp_data["url"],
                data=resp_data["fields"],
                files={"file": f},
            )


@experimental
def put_cost_information(
    context: Union[OpExecutionContext, AssetExecutionContext],
    metric_name: str,
    cost_information: List[Tuple[str, float, str]],
    start: float,
    end: float,
    submit_gql: bool = True,
) -> None:
    chunk_size = _cost_information_chunk_size()

    try:
        import pyarrow as pyarrow  # pylint: disable=import-error

        try:
            upload_cost_information(context, metric_name, cost_information)
        except Exception:
            context.log.warn("Failed to upload cost information to S3.", exc_info=True)
            # if we're not submitting via GQL API as well, raise the exception more loudly
            if not submit_gql:
                raise
    except ImportError:
        context.log.warn(
            "Dagster insights dependencies not installed. In the future, you will need to install dagster-cloud[insights] to use this feature."
        )
        # if we're not submitting via GQL API as well, raise the exception more loudly
        if not submit_gql:
            raise

    # early exit if we're not submitting via GQL API as well
    if not submit_gql:
        return

    cost_info_input = [
        {
            "opaqueId": opaque_id,
            "cost": float(cost),
        }
        for opaque_id, cost, _ in cost_information
    ]

    # Chunk the cost information & keep each opaque id in the same chunk
    # to avoid the cost information for a single asset being split across
    # multiple steps
    for chunk in chunk_by_opaque_id(
        [(opaque_id, cost) for opaque_id, cost, _ in cost_information], chunk_size
    ):
        cost_info_input = [
            {
                "opaqueId": opaque_id,
                "cost": float(cost),
            }
            for opaque_id, cost in chunk
        ]

        result = query_graphql_from_instance(
            context.instance,
            PUT_COST_INFORMATION_MUTATION,
            variables={
                "costInfo": cost_info_input,
                "metricName": metric_name,
                "start": start,
                "end": end,
            },
        )
        if (
            result["submitCostInformationForMetrics"]["__typename"]
            != "CreateOrUpdateMetricsSuccess"
        ):
            raise RuntimeError("Failed to submit cost information", result)


@experimental
def put_metrics(
    context: Union[OpExecutionContext, AssetExecutionContext],
    run_id: str,
    step_key: str,
    metrics: List[DagsterMetric],
    asset_key: Optional[str] = None,
    partition: Optional[str] = None,
) -> Dict[str, Any]:
    """Experimental: Store metrics in the dagster cloud metrics store. This method is useful when you would like to
    store run, asset or asset group materialization metric data to view in the insights UI.

    To associate a metric with an asset, you must provide the asset_key and partition arguments. If you do not provide
    these arguments, the metric will be associated with the job. Often you will want to store both job and asset
    metrics, in which case you should call this method twice, once with the asset_key and partition arguments, and
    once without.

    Currently only supported in Dagster Cloud.

    Args:
        context (Union[OpExecutionContext, AssetExecutionContext]): the execution context used to query Cloud
        run_id (str): id of the dagster run
        step_key (str): key of the step
        code_location_name (str): name of the code location
        repository_name (str): name of the repository
        metrics (List[DagsterMetric]): metrics to store in the dagster metrics store
        asset_key (Optional[str]): key of the asset, if storing an asset metric
        partition (Optional[str]): partition of the asset, if storing an asset metric
    """
    check.list_param(metrics, "metrics", of_type=DagsterMetric)
    check.str_param(run_id, "run_id", run_id)
    check.str_param(step_key, "step_key", step_key)
    check.opt_str_param(asset_key, "asset_key", asset_key)
    check.opt_str_param(partition, "partition", partition)

    url, token = get_url_and_token_from_instance(context.instance)

    transport = RequestsHTTPTransport(
        url=url,
        use_json=True,
        timeout=300,
        headers={"Dagster-Cloud-Api-Token": token},
    )
    Client(transport=transport, fetch_schema_from_transport=True)

    metric_graphql_input = {
        "runId": run_id,
        "stepKey": step_key,
        "assetMetricDefinitions": [],
        "jobMetricDefinitions": [],
    }

    if asset_key is not None:
        metric_graphql_input["assetMetricDefinitions"].append(
            {
                "assetKey": asset_key,
                "partition": partition,
                "metricValues": [
                    {
                        "metricValue": metric_def.metric_value,
                        "metricName": metric_def.metric_name,
                    }
                    for metric_def in metrics
                ],
            }
        )
    else:
        metric_graphql_input["jobMetricDefinitions"].append(
            {
                "metricValues": [
                    {
                        "metricValue": metric_def.metric_value,
                        "metricName": metric_def.metric_name,
                    }
                    for metric_def in metrics
                ],
            }
        )

    return query_graphql_from_instance(
        context.instance,
        PUT_CLOUD_METRICS_MUTATION,
        variables={"metrics": [metric_graphql_input]},
    )



import datetime
import io
import os
import pickle
from typing import Sequence, Tuple, Union

import boto3
import dagster._check as check
import requests
from dagster import InputContext, MemoizableIOManager, OutputContext, io_manager
from dagster._utils import PICKLE_PROTOCOL
from dateutil import parser

ECS_AGENT_IP = "169.254.170.2"


class PickledObjectServerlessIOManager(MemoizableIOManager):
    def __init__(
        self,
        s3_bucket,
        s3_prefix,
    ):
        self._bucket = check.str_param(s3_bucket, "s3_bucket")
        self._s3_prefix = check.str_param(s3_prefix, "s3_prefix")
        self._boto_session, self._boto_session_expiration = self._refresh_boto_session()

    def _refresh_boto_session(self) -> Tuple[boto3.Session, datetime.datetime]:
        # We have to do this whacky way to get credentials to ensure that we get iam role
        # we assigned to the task. If we used the default boto behavior, it could get overriden
        # when users set AWS env vars.
        relative_uri = os.environ["AWS_CONTAINER_CREDENTIALS_RELATIVE_URI"]
        aws_creds = requests.get(f"http://{ECS_AGENT_IP}{relative_uri}").json()
        session = boto3.Session(
            aws_access_key_id=aws_creds["AccessKeyId"],
            aws_secret_access_key=aws_creds["SecretAccessKey"],
            aws_session_token=aws_creds["Token"],
        )
        expiration = parser.parse(aws_creds["Expiration"])
        return session, expiration

    @property
    def _s3(self):
        if self._boto_session_expiration <= datetime.datetime.now(
            self._boto_session_expiration.tzinfo
        ) + datetime.timedelta(minutes=5):
            self._boto_session, self._boto_session_expiration = self._refresh_boto_session()
        return self._boto_session.client("s3", region_name="us-west-2")

    def _get_path(self, context: Union[InputContext, OutputContext]) -> str:
        path: Sequence[str]
        if context.has_asset_key:
            path = context.get_asset_identifier()
        else:
            path = ["storage", *context.get_identifier()]

        return "/".join([self._s3_prefix, *path])

    def has_output(self, context):
        key = self._get_path(context)
        return self._has_object(key)

    def _rm_object(self, key):
        check.str_param(key, "key")
        check.param_invariant(len(key) > 0, "key")

        # delete_object wont fail even if the item has been deleted.
        self._s3.delete_object(Bucket=self._bucket, Key=key)

    def _has_object(self, key):
        check.str_param(key, "key")
        check.param_invariant(len(key) > 0, "key")

        found_object = False

        try:
            self._s3.get_object(Bucket=self._bucket, Key=key)
            found_object = True
        except self._s3.exceptions.NoSuchKey:
            found_object = False

        return found_object

    def load_input(self, context):
        if context.dagster_type.typing_type == type(None):
            return None

        key = self._get_path(context)
        check.invariant(self._has_object(key), "Input not found. It may have expired.")
        obj = pickle.loads(self._s3.get_object(Bucket=self._bucket, Key=key)["Body"].read())

        return obj

    def handle_output(self, context, obj):
        if context.dagster_type.typing_type == type(None):
            check.invariant(
                obj is None,
                "Output had Nothing type or 'None' annotation, but handle_output received"
                f" value that was not None and was of type {type(obj)}.",
            )
            return None

        key = self._get_path(context)

        if self._has_object(key):
            self._rm_object(key)

        pickled_obj = pickle.dumps(obj, PICKLE_PROTOCOL)
        pickled_obj_bytes = io.BytesIO(pickled_obj)
        self._s3.upload_fileobj(
            pickled_obj_bytes,
            self._bucket,
            key,
        )


@io_manager
def serverless_io_manager(init_context):
    bucket = os.getenv("DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET")
    prefix = os.getenv("DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX")
    check.invariant(
        bucket and prefix,
        "The serverless_io_manager is only supported when running on Dagster Cloud Serverless."
        " DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET or"
        " DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX not found.",
    )

    deployment_name = init_context.instance.deployment_name

    return PickledObjectServerlessIOManager(
        bucket, s3_prefix=f"{prefix}/io_storage/{deployment_name}"
    )



import os
from typing import Optional, Sequence, Union, cast

from dagster import (
    _check as check,
    asset_check,
)
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.events import CoercibleToAssetKey
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.errors import (
    DagsterError,
    DagsterInvariantViolationError,
)
from dagster._core.execution.context.compute import AssetExecutionContext
from dagster._core.instance import DagsterInstance
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport

from dagster_cloud import DagsterCloudAgentInstance

from .mutation import ANOMALY_DETECTION_INFERENCE_MUTATION
from .types import (
    AnomalyDetectionModelParams,
    BetaFreshnessAnomalyDetectionParams,
)

DEFAULT_MODEL_PARAMS = BetaFreshnessAnomalyDetectionParams(sensitivity=0.1)


class DagsterCloudAnomalyDetectionFailed(DagsterError):
    """Raised when an anomaly detection check fails host-side."""


def _build_check_for_asset(
    asset: Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset],
    params: AnomalyDetectionModelParams,
) -> AssetChecksDefinition:
    @asset_check(
        asset=asset,
        description=f"Detects anomalies in the freshness of the asset using model {params.model_version.value.lower()}.",
        name="freshness_anomaly_detection",
    )
    def the_check(context: AssetExecutionContext) -> AssetCheckResult:
        if not _is_agent_instance(context.instance):
            raise DagsterInvariantViolationError(
                f"This anomaly detection check is not being launched from a dagster agent. "
                "Anomaly detection is only available for dagster cloud deployments."
                f"Instance type: {type(context.instance)}."
            )
        instance = cast(DagsterCloudAgentInstance, context.instance)
        transport = RequestsHTTPTransport(
            url=os.getenv("DAGSTER_METRICS_DAGIT_URL", f"{instance.dagit_url}graphql"),
            use_json=True,
            timeout=300,
            headers={"Dagster-Cloud-Api-Token": instance.dagster_cloud_agent_token},
        )
        client = Client(transport=transport, fetch_schema_from_transport=True)
        asset_key = context.asset_checks_def.asset_key
        asset = context.job_def.asset_layer.assets_defs_by_key.get(
            asset_key
        ) or context.job_def.asset_layer.source_assets_by_key.get(asset_key)
        if asset is None:
            raise Exception(f"Could not find targeted asset {asset_key.to_string()}.")
        result = client.execute(
            gql(ANOMALY_DETECTION_INFERENCE_MUTATION),
            {
                "modelVersion": params.model_version.value,
                "params": {
                    **dict(params),
                    "asset_key_user_string": asset_key.to_user_string(),
                },
            },
        )
        if result["anomalyDetectionInference"]["__typename"] != "AnomalyDetectionSuccess":
            raise DagsterCloudAnomalyDetectionFailed(
                f"Anomaly detection failed: {result['anomalyDetectionInference']['message']}"
            )
        response = result["anomalyDetectionInference"]["response"]
        overdue_seconds = check.float_param(response["overdue_seconds"], "overdue_seconds")
        expected_event_timestamp = response["overdue_deadline_timestamp"]
        model_training_range_start = response["model_training_range_start_timestamp"]
        model_training_range_end = response["model_training_range_end_timestamp"]
        metadata = {
            "model_params": {**params.as_metadata},
            "model_version": params.model_version.value,
            "model_training_range_start_timestamp": model_training_range_start,
            "model_training_range_end_timestamp": model_training_range_end,
            "overdue_deadline_timestamp": expected_event_timestamp,
        }
        if overdue_seconds > 0:
            metadata["overdue_minutes"] = overdue_seconds / 60
            return AssetCheckResult(
                passed=False,
                severity=AssetCheckSeverity.WARN,
                metadata=metadata,
            )
        else:
            return AssetCheckResult(passed=True, metadata=metadata)

    return the_check


def build_anomaly_detection_freshness_checks(
    *,
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
    params: Optional[AnomalyDetectionModelParams],
) -> Sequence[AssetChecksDefinition]:
    """Builds a list of asset checks which utilize anomaly detection algorithms to
    determine the freshness of data.

    Args:
        assets (Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]]): The assets to construct checks for. For each passed in
            asset, there will be a corresponding constructed `AssetChecksDefinition`.
        params (AnomalyDetectionModelParams): The parameters to use for the model. The parameterization corresponds to the model used.

    Returns:
        Sequence[AssetChecksDefinition]: A list of `AssetChecksDefinition` objects, each corresponding to an asset in the `assets` parameter.

    Examples:
        .. code-block:: python

            from dagster_cloud import build_anomaly_detection_freshness_checks, AnomalyDetectionModel, BetaFreshnessAnomalyDetectionParams

            checks = build_anomaly_detection_freshness_checks(
                assets=[AssetKey("foo_asset"), AssetKey("foo_asset")],
                params=BetaFreshnessAnomalyDetectionParams(sensitivity=0.1),
            )
    """
    params = check.opt_inst_param(
        params, "params", AnomalyDetectionModelParams, DEFAULT_MODEL_PARAMS
    )
    return [_build_check_for_asset(asset, params) for asset in assets]


def _is_agent_instance(instance: DagsterInstance) -> bool:
    if hasattr(instance, "dagster_cloud_agent_token") and hasattr(instance, "dagit_url"):
        return True
    return False



import os
from typing import Any, Dict, Generator, List

import dagster._check as check
from dagster import AssetObservation, OpExecutionContext
from dagster._annotations import experimental
from gql import Client, gql
from gql.transport.exceptions import TransportError
from gql.transport.requests import RequestsHTTPTransport

from ..dagster_insights.errors import DagsterInsightsError
from ..dagster_insights.metrics_utils import (
    PUT_CLOUD_METRICS_MUTATION,
    DagsterMetric as DagsterMetric,
    put_metrics as put_metrics,
)
from ..instance import DagsterCloudAgentInstance


def _chunks(chunk_list: List[Any], length: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(chunk_list), length):
        yield chunk_list[i : i + length]


@experimental
def put_context_metrics(
    context: OpExecutionContext,
    metrics: List[DagsterMetric],
) -> None:
    """Experimental: Store metrics in the dagster cloud metrics store. This method is useful when you would like to
    store run, asset or asset group materialization metric data to view in the insights UI.

    Currently only supported in Dagster Cloud

    Args:
        context (OpExecutionContext): the execution context used for asset materialization
        metrics (List[DagsterMetric]): metrics to store in the dagster metrics store
    """
    check.list_param(metrics, "metrics", of_type=DagsterMetric)
    check.inst_param(context, "context", OpExecutionContext)
    if not isinstance(context.instance, DagsterCloudAgentInstance):
        context.log.info("Dagster instance is not a DagsterCloudAgentInstance, skipping metrics")
        return

    transport = RequestsHTTPTransport(
        url=os.getenv("DAGSTER_METRICS_DAGIT_URL", f"{context.instance.dagit_url}graphql"),
        use_json=True,
        timeout=300,
        headers={"Dagster-Cloud-Api-Token": context.instance.dagster_cloud_agent_token},
    )
    client = Client(transport=transport, fetch_schema_from_transport=True)

    metric_graphql_inputs: List[Dict[str, Any]] = []

    if context.dagster_run.external_job_origin is None:
        raise DagsterInsightsError("dagster run for this context has not started yet")

    if context.has_assets_def:
        for selected_asset_keys in _chunks(list(context.selected_asset_keys), 5):
            metric_graphql_inputs.append(
                {
                    "runId": context.run_id,
                    "stepKey": context.get_step_execution_context().step.key,
                    "codeLocationName": context.dagster_run.external_job_origin.location_name,
                    "repositoryName": (
                        context.dagster_run.external_job_origin.external_repository_origin.repository_name
                    ),
                    "assetMetricDefinitions": [
                        {
                            "assetKey": selected_asset_key.to_python_identifier(),
                            "assetGroup": context.assets_def.group_names_by_key.get(
                                selected_asset_key, None
                            ),
                            "partition": (
                                context.partition_key if context.has_partition_key else None
                            ),
                            "metricValues": [
                                {
                                    "metricValue": metric_def.metric_value,
                                    "metricName": metric_def.metric_name,
                                }
                                for metric_def in metrics
                            ],
                        }
                        for selected_asset_key in selected_asset_keys
                    ],
                }
            )
    else:
        metric_graphql_inputs.append(
            {
                "runId": context.run_id,
                "stepKey": context.get_step_execution_context().step.key,
                "codeLocationName": context.dagster_run.external_job_origin.location_name,
                "repositoryName": (
                    context.dagster_run.external_job_origin.external_repository_origin.repository_name
                ),
                "jobMetricDefinitions": [
                    {
                        "metricValues": [
                            {
                                "metricValue": metric_def.metric_value,
                                "metricName": metric_def.metric_name,
                            }
                            for metric_def in metrics
                        ],
                    }
                ],
            }
        )
    for metric_graphql_inputs in _chunks(metric_graphql_inputs, 5):
        try:
            result = client.execute(
                gql(PUT_CLOUD_METRICS_MUTATION),
                variable_values={"metrics": metric_graphql_inputs},
            )
            if (
                result.get("createOrUpdateMetrics", {}).get("__typename")
                != "CreateOrUpdateMetricsSuccess"
            ):
                context.log.error(
                    "Failed to store metrics with error"
                    f" {result.get('createOrUpdateMetrics', {}).get('message')}"
                )
        except (TransportError, TypeError, KeyError) as exc:
            context.log.error(f"Failed to store metrics with error {exc}")
            return

    context.log.info("Successfully stored metrics")


@experimental
def store_dbt_adapter_metrics(
    context: OpExecutionContext,
    manifest: Dict[Any, Any],
    run_results: Dict[Any, Any],
) -> Generator[AssetObservation, Any, Any]:
    """Experimental: Store dbt adapter response metrics in the dagster cloud metrics store. This method is useful when you would like to store dbt adapter response metrics to view in the insights UI.

    Currently only supported in Dagster Cloud

    Args:
        context (OpExecutionContext): the execution context used for asset materialization
        manifest (Dict[Any, Any]): the manifest of the dbt command
        run_results (Dict[Any, Any]): the run results of the dbt command
    """
    check.inst_param(context, "context", OpExecutionContext)
    check.dict_param(manifest, "manifest")
    check.dict_param(run_results, "run_results")

    if not isinstance(context.instance, DagsterCloudAgentInstance):
        context.log.info("Dagster instance is not a DagsterCloudAgentInstance, skipping metrics")
        return

    transport = RequestsHTTPTransport(
        url=os.getenv("DAGSTER_METRICS_DAGIT_URL", f"{context.instance.dagit_url}graphql"),
        use_json=True,
        timeout=300,
        headers={"Dagster-Cloud-Api-Token": context.instance.dagster_cloud_agent_token},
    )
    client = Client(transport=transport, fetch_schema_from_transport=True)

    if context.dagster_run.external_job_origin is None:
        raise DagsterInsightsError("dagster run for this context has not started yet")
    assetMetricDefinitions = []
    for result in run_results["results"]:
        node = manifest["nodes"][result["unique_id"]]
        metric_values = []
        assetKey = next(
            iter(
                filter(
                    lambda asset_key: asset_key.path[-1] == node["name"],
                    context.selected_asset_keys,
                )
            ),
            None,
        )

        if not assetKey:
            continue

        for adapter_response_key in result["adapter_response"]:
            if adapter_response_key in ["_message", "code"]:
                continue
            if adapter_response_key == "query_id":
                yield AssetObservation(
                    asset_key=assetKey,
                    metadata={"query_id": result["adapter_response"][adapter_response_key]},
                )
            if isinstance(result["adapter_response"][adapter_response_key], float) or isinstance(
                result["adapter_response"][adapter_response_key], int
            ):
                metric_values.append(
                    {
                        "metricValue": result["adapter_response"][adapter_response_key],
                        "metricName": adapter_response_key,
                    }
                )
        partition = None
        try:
            partition = context.asset_partition_key_for_output(assetKey)
        except check.CheckError:
            pass

        assetMetricDefinitions.append(
            {
                "assetKey": assetKey.to_python_identifier(),
                "assetGroup": context.assets_def.group_names_by_key.get(assetKey, None),
                "partition": partition,
                "metricValues": metric_values,
            }
        )
    metric_graphql_inputs = []
    for assetMetricDefinitions in _chunks(assetMetricDefinitions, 5):
        metric_graphql_inputs.append(
            {
                "runId": context.run_id,
                "stepKey": context.get_step_execution_context().step.key,
                "codeLocationName": context.dagster_run.external_job_origin.location_name,
                "repositoryName": (
                    context.dagster_run.external_job_origin.external_repository_origin.repository_name
                ),
                "assetMetricDefinitions": assetMetricDefinitions,
            }
        )
    for metric_graphql_inputs in _chunks(metric_graphql_inputs, 5):
        try:
            result = client.execute(
                gql(PUT_CLOUD_METRICS_MUTATION),
                variable_values={"metrics": metric_graphql_inputs},
            )
            if (
                result.get("createOrUpdateMetrics", {}).get("__typename")
                != "CreateOrUpdateMetricsSuccess"
            ):
                context.log.error(
                    "Failed to store metrics with error"
                    f" {result.get('createOrUpdateMetrics', {}).get('message')}"
                )
        except (TransportError, TypeError, KeyError) as exc:
            context.log.error(f"Failed to store metrics with error {exc}")
            return
    context.log.info("Successfully stored metrics")



from typing import Any, Dict, List, Optional, cast

from dagster import (
    Enum as DagsterEnum,
    EnumValue,
    Field,
    Map,
    Noneable,
    Permissive,
    Selector,
    Shape,
    _check as check,
)
from dagster._config import EvaluationError, StringSource, validate_config

from .docker import SHARED_DOCKER_CONFIG
from .ecs import (
    ECS_CONTAINER_CONTEXT_CONFIG as ECS_CONTAINER_CONTEXT_CONFIG,
    SHARED_ECS_CONFIG as SHARED_ECS_CONFIG,
)
from .kubernetes import SHARED_K8S_CONFIG


def validate_workspace_location(workspace_location) -> Optional[List[str]]:
    """Processes a single workspace location config. Returns a list of error
    messages if any.
    """
    validation = validate_config(LOCATION_CONFIG_SCHEMA, workspace_location)
    return [error.message for error in validation.errors or []]


def validate_workspace_config(workspace_config) -> Optional[List[str]]:
    """Processes an entire workspace location config. Returns a list of
    error messages, if any.
    """
    validation = validate_config(WORKSPACE_CONFIG_SCHEMA, workspace_config)
    return [error.message for error in validation.errors or []]


def process_workspace_config(workspace_config) -> Dict[str, Any]:
    """Checks a workspace config, erroring if any mismatches with config
    and migrating an input in the legacy workspace config format to the
    modern format, returning the validated input.
    """
    check.dict_param(workspace_config, "workspace_config")

    # Check if using the legacy format, see below for details
    if isinstance(workspace_config.get("locations"), dict):
        check.is_dict(workspace_config.get("locations"))

        validation = validate_config(LEGACY_WORKSPACE_CONFIG_SCHEMA, workspace_config)
        check.invariant(
            validation.success,
            ", ".join([error.message for error in validation.errors or []]),
        )

        locations = workspace_config["locations"].values()

        # Manually validate that only one of python_file, package_name, module_name specified
        # modern format will check this automatically as it uses a Selector ConfigType
        for config in locations:
            python_file = config.get("python_file")
            package_name = config.get("package_name")
            module_name = config.get("module_name")
            check.invariant(
                len([val for val in [python_file, package_name, module_name] if val]) == 1,
                "Must supply exactly one of a file name, a package name, or a module name",
            )

        # Convert legacy formatted locations to modern format
        updated_locations = []
        for name, location in workspace_config["locations"].items():
            new_location = {
                k: v
                for k, v in location.items()
                if k not in ("python_file", "package_name", "module_name")
            }
            new_location["code_source"] = {}
            if "python_file" in location:
                new_location["code_source"]["python_file"] = location["python_file"]
            elif "package_name" in location:
                new_location["code_source"]["package_name"] = location["package_name"]
            elif "module_name" in location:
                new_location["code_source"]["module_name"] = location["module_name"]

            new_location["location_name"] = name
            updated_locations.append(new_location)
        return {"locations": updated_locations}
    else:
        check.is_list(workspace_config.get("locations"))

        validation = validate_config(WORKSPACE_CONFIG_SCHEMA, workspace_config)
        check.invariant(
            validation.success,
            ", ".join([error.message for error in cast(List[EvaluationError], validation.errors)]),
        )
        return workspace_config


K8S_CONFIG_FIELDS = {
    **SHARED_K8S_CONFIG,
    "run_k8s_config": Field(
        Shape(
            {
                "container_config": Permissive(),
                "pod_template_spec_metadata": Permissive(),
                "pod_spec_config": Permissive(),
                "job_config": Permissive(),
                "job_metadata": Permissive(),
                "job_spec_config": Permissive(),
                "merge_behavior": Field(
                    DagsterEnum(
                        "K8sConfigMergeBehavior",
                        [
                            EnumValue("SHALLOW"),
                            EnumValue("DEEP"),
                        ],
                    ),
                    is_required=False,
                ),
            }
        ),
        is_required=False,
        description="Raw Kubernetes configuration for launched runs.",
    ),
    "server_k8s_config": Field(
        Shape(
            {
                "container_config": Permissive(),
                "pod_spec_config": Permissive(),
                "pod_template_spec_metadata": Permissive(),
                "merge_behavior": Field(
                    DagsterEnum(
                        "K8sConfigMergeBehavior",
                        [
                            EnumValue("SHALLOW"),
                            EnumValue("DEEP"),
                        ],
                    ),
                    is_required=False,
                ),
            }
        ),
        is_required=False,
        description="Raw Kubernetes configuration for launched code servers.",
    ),
}


CONFIG_SCHEMA_FIELDS = {
    # Only used by the new workspace.yaml format, legacy format has the name as a key
    "location_name": Field(config=str, is_required=True, description="Location name"),
    "image": Field(
        config=Noneable(str),
        is_required=False,
        description="Docker image, for use with containerized agents.",
    ),
    "pex_metadata": Field(
        Shape(
            fields={
                "pex_tag": Field(
                    config=str,
                    is_required=True,
                    description="Identifies which Pex files to use (internal only).",
                ),
                "python_version": Field(
                    config=str,
                    is_required=False,
                    description=(
                        "Determines which base image to use for pex, alternative to specifying the"
                        " top level image."
                    ),
                ),
            }
        ),
        is_required=False,
    ),
    "code_source": Field(
        Selector(
            {
                "python_file": Field(
                    config=str,
                    description="Python file containing the target Dagster repository.",
                ),
                "package_name": Field(
                    config=str,
                    description="Python package containing the target Dagster repository.",
                ),
                "module_name": Field(
                    config=str,
                    description="Python module containing the target Dagster repository.",
                ),
            },
        ),
        description="Python entry point for the code location.",
    ),
    "working_directory": Field(
        config=str,
        is_required=False,
        description=(
            "Working directory to use for importing Python modules when loading the repository."
        ),
    ),
    "executable_path": Field(
        config=str,
        is_required=False,
        description=(
            "Path to reach the executable to use for the Python environment to load the repository."
            " Defaults to the installed dagster command-line entry point."
        ),
    ),
    "attribute": Field(
        config=str,
        is_required=False,
        description=(
            "Specifies either a repository or a function that returns a repository. Can be used"
            " when the code contains multiple repositories but only one should be included."
        ),
    ),
    "git": Field(
        Shape(
            fields={
                "commit_hash": Field(
                    config=str,
                    is_required=False,
                    description="Indicates the commit sha associated with this location.",
                ),
                "url": Field(
                    config=str,
                    is_required=False,
                    description="Specifies a source code reference link for this location.",
                ),
            }
        ),
        is_required=False,
    ),
    "container_context": Field(
        Shape(
            fields={
                "k8s": Field(Shape(K8S_CONFIG_FIELDS), is_required=False),
                "docker": Field(Shape(SHARED_DOCKER_CONFIG), is_required=False),
                "ecs": Field(Shape(ECS_CONTAINER_CONTEXT_CONFIG), is_required=False),
            },
        ),
        description="Metadata for specific compute environments",
        is_required=False,
    ),
    "agent_queue": Field(
        config=str,
        is_required=False,
        description="Locations that specify an agent queue will only have their requests handled by agents configured to read from a matching queue. By default, requests are placed on a default queue that's handled by all agents.",
    ),
}


# The legacy format is structured as
# locations:
#   my_location_name:
#     image: abc
#     python_file: repo.py
#
# Here, we drop the fields not included in the legacy format, including location name (which is the
# dict key in the legacy format) and the code_source field (whose children are top-level values).
LEGACY_CONFIG_SCHEMA_FIELDS = {
    **{k: v for k, v in CONFIG_SCHEMA_FIELDS.items() if k not in ("location_name", "code_source")},
    "python_file": Field(
        config=StringSource,
        is_required=False,
        description="Python file containing the target Dagster repository.",
    ),
    "package_name": Field(
        config=StringSource,
        is_required=False,
        description="Python package containing the target Dagster repository.",
    ),
    "module_name": Field(
        config=StringSource,
        is_required=False,
        description="Python module containing the target Dagster repository.",
    ),
}
LEGACY_LOCATION_CONFIG_SCHEMA = Shape(fields=LEGACY_CONFIG_SCHEMA_FIELDS)
LEGACY_NAMED_LOCATIONS_CONFIG_SCHEMA = Map(
    str, LEGACY_LOCATION_CONFIG_SCHEMA, key_label_name="location_name"
)
LEGACY_WORKSPACE_CONFIG_SCHEMA = Shape(
    fields={
        "locations": Field(
            LEGACY_NAMED_LOCATIONS_CONFIG_SCHEMA,
            description="Mapping from location names to location details.",
        )
    }
)

# The modern format is structured as
# locations:
#   - location_name: my_location_name
#     image: abc
#     code_source:
#       python_file: repo.py
LOCATION_CONFIG_SCHEMA = Shape(fields=CONFIG_SCHEMA_FIELDS)
WORKSPACE_CONFIG_SCHEMA = Shape(fields={"locations": [LOCATION_CONFIG_SCHEMA]})



from typing import Optional, Tuple, Union

import dagster._check as check
from dagster import (
    AssetExecutionContext,
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    DagsterInvariantViolationError,
    OpExecutionContext,
    Output,
)
from dagster._core.errors import DagsterInvalidPropertyError


def get_current_context_and_asset_key() -> (
    Tuple[Union[OpExecutionContext, AssetExecutionContext], Optional[AssetKey]]
):
    asset_key = None
    try:
        context = AssetExecutionContext.get()
        if len(context.assets_def.keys_by_output_name.keys()) == 1:
            asset_key = context.asset_key
    except (DagsterInvalidPropertyError, DagsterInvariantViolationError):
        context = OpExecutionContext.get()

    return check.not_none(context), asset_key


def get_asset_key_for_output(
    context: Union[OpExecutionContext, AssetExecutionContext], output_name: str
) -> Optional[AssetKey]:
    asset_info = context.job_def.asset_layer.asset_info_for_output(
        node_handle=context.op_handle, output_name=output_name
    )
    if asset_info is None:
        return None
    return asset_info.key


def extract_asset_info_from_event(context, dagster_event, record_observation_usage):
    if isinstance(dagster_event, AssetMaterialization):
        return dagster_event.asset_key, dagster_event.partition

    if isinstance(dagster_event, AssetObservation) and record_observation_usage:
        return dagster_event.asset_key, dagster_event.partition

    if isinstance(dagster_event, AssetObservation):
        return None, None

    if isinstance(dagster_event, Output):
        asset_key = get_asset_key_for_output(context, dagster_event.output_name)
        partition_key = None
        if asset_key and context._step_execution_context.has_asset_partitions_for_output(  # noqa: SLF001
            dagster_event.output_name
        ):
            # We associate cost with the first partition key in the case that an output
            # maps to multiple partitions. This is a temporary solution, but partition key
            # is not used in Insights at the moment.
            # TODO: Find a long-term solution for this
            partition_key = next(
                iter(context.asset_partition_keys_for_output(dagster_event.output_name))
            )

        return asset_key, partition_key

    return None, None



import json
import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

import yaml
from dagster._core.errors import DagsterHomeNotSetError
from dagster._utils.interrupts import capture_interrupts
from dagster._utils.log import default_date_format_string, default_format_string
from dagster._utils.merger import deep_merge_dicts
from dagster._utils.yaml_utils import load_yaml_from_globs
from dagster_cloud_cli import ui
from typer import Argument, Option, Typer

from dagster_cloud.agent.dagster_cloud_agent import DagsterCloudAgent
from dagster_cloud.instance import DagsterCloudAgentInstance

app = Typer(help="Interact with the Dagster Cloud agent.")


def agent_home_exception():
    dagster_home_loc = (
        f"No Dagster config provided in specified directory {os.getenv('DAGSTER_HOME')}. "
        "You must specify the location of a directory containing a dagster.yaml "
        "file as a parameter or by setting the DAGSTER_HOME environment variable."
        if os.getenv("DAGSTER_HOME")
        else (
            "No directory provided or DAGSTER_CLOUD environment variable set. "
            "You must supply the location of a directory containing a dagster.yaml "
            "file as a parameter or by setting the DAGSTER_HOME environment variable."
        )
    )
    return ui.error(f"No Dagster config found.\n\n{dagster_home_loc}")


def run_local_agent(agent_logging_config_path):
    try:
        with DagsterCloudAgentInstance.get() as instance:
            logging.basicConfig(
                level=logging.INFO,
                format=default_format_string(),
                datefmt=default_date_format_string(),
                handlers=[logging.StreamHandler()],
            )

            if agent_logging_config_path:
                agent_logging_config = {
                    "version": 1,
                    "disable_existing_loggers": False,
                    **load_yaml_from_globs(str(agent_logging_config_path)),
                }
                logging.config.dictConfig(agent_logging_config)

            user_code_launcher = instance.user_code_launcher
            user_code_launcher.start()

            with DagsterCloudAgent() as agent:
                agent.run_loop(instance, user_code_launcher, agent_uuid=instance.instance_uuid)
    except DagsterHomeNotSetError:
        raise agent_home_exception()


def run_local_agent_in_environment(
    dagster_home: Optional[Path], agent_logging_config_path: Optional[Path]
):
    with capture_interrupts():
        old_env = None
        try:
            old_env = dict(os.environ)
            if dagster_home:
                os.environ["DAGSTER_HOME"] = str(dagster_home.resolve())
            run_local_agent(agent_logging_config_path)
        finally:
            os.environ.clear()
            if old_env is not None:
                os.environ.update(old_env)


def run_local_agent_in_temp_environment(
    agent_token: str,
    deployment: str,
    agent_label: Optional[str],
    instance_config: Optional[str],
    user_code_launcher_module: Optional[str],
    user_code_launcher_class: Optional[str],
    user_code_launcher_config: Optional[str],
):
    config = {
        "instance_class": {
            "module": "dagster_cloud.instance",
            "class": "DagsterCloudAgentInstance",
        },
        "dagster_cloud_api": {},
        "user_code_launcher": {
            "module": "dagster_cloud.workspace.user_code_launcher",
            "class": "ProcessUserCodeLauncher",
        },
    }
    if instance_config:
        parsed = json.loads(instance_config)
        config = deep_merge_dicts(config, parsed)

    if agent_token:
        config["dagster_cloud_api"]["agent_token"] = agent_token
    if deployment:
        config["dagster_cloud_api"]["deployment"] = deployment
    if agent_label:
        config["dagster_cloud_api"]["agent_label"] = agent_label
    if user_code_launcher_module:
        config["user_code_launcher"]["module"] = user_code_launcher_module
    if user_code_launcher_class:
        config["user_code_launcher"]["class"] = user_code_launcher_class
    if user_code_launcher_config:
        try:
            config["user_code_launcher"]["config"] = json.loads(user_code_launcher_config)
        except json.JSONDecodeError as e:
            raise ui.error(f"Invalid User Code Launcher config JSON:\n{e}")

    with TemporaryDirectory() as d:
        with open(os.path.join(d, "dagster.yaml"), "w", encoding="utf8") as f:
            f.write(yaml.dump(config))
        run_local_agent_in_environment(Path(d), None)


@app.command(
    help=(
        "Runs the Dagster Cloud agent. The agent can either be run ephemerally by specifying an"
        " agent token and deployment name as CLI options, or the agent can pull its config from a"
        " dagster.yaml file. To use a dagster.yaml file, either pass a directory containing the"
        " file as a CLI argument or set the DAGSTER_HOME environment variable."
    ),
    short_help="Run the Dagster Cloud agent.",
)
def run(
    dagster_home: Optional[Path] = Argument(None),
    agent_token: str = Option(
        None, "--agent-token", "-a", help="Agent token, if running ephemerally."
    ),
    deployment: str = Option(
        None, "--deployment", "-d", help="Deployment, if running ephemerally."
    ),
    agent_label: str = Option(
        None, "--agent-label", "-l", help="Optional agent label, if running ephemerally."
    ),
    user_code_launcher: str = Option(
        None,
        "--user-code-launcher",
        help="User Code Launcher to use. Defaults to the local Process User Code Launcher.",
        hidden=True,
    ),
    instance_config: str = Option(
        None,
        "--config",
        help="Dagster instance config, in JSON format.",
    ),
    user_code_launcher_config: str = Option(
        None,
        "--user-code-launcher-config",
        help="Config to supply the User Code Launcher, in JSON format.",
        hidden=True,
    ),
    agent_logging_config_path: Optional[Path] = Option(
        None,
        "--agent-logging-config-path",
        help=(
            "Yaml file with logging config for the agent process that can be passed into"
            " logging.dictConfig"
        ),
        exists=True,
    ),
):
    if (
        agent_token
        or deployment
        or agent_label
        or instance_config
        or user_code_launcher_config
        or user_code_launcher
    ):
        if not instance_config and (not agent_token or not deployment):
            raise ui.error("To run ephemerally, must supply both an agent token and a deployment.")
        if dagster_home:
            raise ui.error("Cannot supply both a dagster home directory and ephemeral parameters.")

        if user_code_launcher:
            user_code_launcher_module = ""
            user_code_launcher_class = user_code_launcher
            if "." in user_code_launcher:
                user_code_launcher_module, user_code_launcher_class = user_code_launcher.rsplit(
                    ".", 1
                )
        else:
            user_code_launcher_module = None
            user_code_launcher_class = None

        run_local_agent_in_temp_environment(
            agent_token,
            deployment,
            agent_label,
            instance_config,
            user_code_launcher_module,
            user_code_launcher_class,
            user_code_launcher_config,
        )
    else:
        run_local_agent_in_environment(dagster_home, agent_logging_config_path)



# @generated

# This file was generated by running `python -m dagster_cloud.pex.grpc.compile`
# Do not edit this file directly, and do not attempt to recompile it using
# grpc_tools.protoc directly, as several changes must be made to the raw output

# ruff: noqa: SLF001

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: multi_pex_api.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13multi_pex_api.proto\x12\rmulti_pex_api\"\x07\n\x05\x45mpty\"8\n\x16\x43reatePexServerRequest\x12\x1e\n\x16\x63reate_pex_server_args\x18\x01 \x01(\t\":\n\x14\x43reatePexServerReply\x12\"\n\x1a\x63reate_pex_server_response\x18\x01 \x01(\t\"4\n\x14GetPexServersRequest\x12\x1c\n\x14get_pex_servers_args\x18\x01 \x01(\t\"6\n\x12GetPexServersReply\x12 \n\x18get_pex_servers_response\x18\x01 \x01(\t\"C\n\x1bGetCrashedPexServersRequest\x12$\n\x1cget_crashed_pex_servers_args\x18\x01 \x01(\t\"E\n\x19GetCrashedPexServersReply\x12(\n get_crashed_pex_servers_response\x18\x01 \x01(\t\"<\n\x18ShutdownPexServerRequest\x12 \n\x18shutdown_pex_server_args\x18\x01 \x01(\t\">\n\x16ShutdownPexServerReply\x12$\n\x1cshutdown_pex_server_response\x18\x01 \x01(\t\"\x1b\n\x0bPingRequest\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t\"\x19\n\tPingReply\x12\x0c\n\x04\x65\x63ho\x18\x01 \x01(\t2\xe0\x03\n\x0bMultiPexApi\x12_\n\x0f\x43reatePexServer\x12%.multi_pex_api.CreatePexServerRequest\x1a#.multi_pex_api.CreatePexServerReply\"\x00\x12Y\n\rGetPexServers\x12#.multi_pex_api.GetPexServersRequest\x1a!.multi_pex_api.GetPexServersReply\"\x00\x12n\n\x14GetCrashedPexServers\x12*.multi_pex_api.GetCrashedPexServersRequest\x1a(.multi_pex_api.GetCrashedPexServersReply\"\x00\x12\x65\n\x11ShutdownPexServer\x12\'.multi_pex_api.ShutdownPexServerRequest\x1a%.multi_pex_api.ShutdownPexServerReply\"\x00\x12>\n\x04Ping\x12\x1a.multi_pex_api.PingRequest\x1a\x18.multi_pex_api.PingReply\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'multi_pex_api_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EMPTY']._serialized_start=38
  _globals['_EMPTY']._serialized_end=45
  _globals['_CREATEPEXSERVERREQUEST']._serialized_start=47
  _globals['_CREATEPEXSERVERREQUEST']._serialized_end=103
  _globals['_CREATEPEXSERVERREPLY']._serialized_start=105
  _globals['_CREATEPEXSERVERREPLY']._serialized_end=163
  _globals['_GETPEXSERVERSREQUEST']._serialized_start=165
  _globals['_GETPEXSERVERSREQUEST']._serialized_end=217
  _globals['_GETPEXSERVERSREPLY']._serialized_start=219
  _globals['_GETPEXSERVERSREPLY']._serialized_end=273
  _globals['_GETCRASHEDPEXSERVERSREQUEST']._serialized_start=275
  _globals['_GETCRASHEDPEXSERVERSREQUEST']._serialized_end=342
  _globals['_GETCRASHEDPEXSERVERSREPLY']._serialized_start=344
  _globals['_GETCRASHEDPEXSERVERSREPLY']._serialized_end=413
  _globals['_SHUTDOWNPEXSERVERREQUEST']._serialized_start=415
  _globals['_SHUTDOWNPEXSERVERREQUEST']._serialized_end=475
  _globals['_SHUTDOWNPEXSERVERREPLY']._serialized_start=477
  _globals['_SHUTDOWNPEXSERVERREPLY']._serialized_end=539
  _globals['_PINGREQUEST']._serialized_start=541
  _globals['_PINGREQUEST']._serialized_end=568
  _globals['_PINGREPLY']._serialized_start=570
  _globals['_PINGREPLY']._serialized_end=595
  _globals['_MULTIPEXAPI']._serialized_start=598
  _globals['_MULTIPEXAPI']._serialized_end=1078
# @@protoc_insertion_point(module_scope)



import json
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
)

import dagster._check as check
from dagster import (
    AssetKey,
    AssetsDefinition,
    DagsterEventType,
    DagsterInstance,
    EventLogRecord,
    EventRecordsFilter,
    ScheduleDefinition,
)

from .snowflake_utils import OPAQUE_ID_METADATA_KEY_PREFIX, OPAQUE_ID_SQL_SIGIL

if TYPE_CHECKING:
    from dagster_snowflake import SnowflakeConnection

EVENT_RECORDS_LIMIT = 10000


@dataclass(frozen=True, eq=True)
class AssetMaterializationId:
    run_id: str
    asset_key: AssetKey
    partition: Optional[str]
    step_key: str


def get_opaque_ids_to_assets(
    instance: DagsterInstance,
    min_datetime: datetime,
    max_datetime: datetime,
) -> Iterable[Tuple[AssetMaterializationId, str]]:
    """Given access to the Dagster GraphQL API, queries for all asset materializations that occurred
    in the provided time range and returns a mapping from AssetMaterializationId to opaque
    Snowflake query ID.
    """
    records: Sequence[EventLogRecord] = []
    last_response = []
    first = True

    while len(last_response) == EVENT_RECORDS_LIMIT or first:
        last_response = instance.get_event_records(
            EventRecordsFilter(
                event_type=DagsterEventType.ASSET_OBSERVATION,
                before_timestamp=max_datetime.timestamp(),
                after_timestamp=min_datetime.timestamp(),
                after_cursor=last_response[-1].storage_id if last_response else None,
            ),
            limit=EVENT_RECORDS_LIMIT,
        )
        first = False
        records.extend(last_response)

    for record in records:
        asset_key = record.asset_key
        if asset_key is None:
            # should never happen, but we must appease the type checker
            raise RuntimeError("asset_key is None, which should never happen")
        run_id = check.not_none(record.run_id)
        partition = record.partition_key
        step_key = check.not_none(record.event_log_entry.step_key)
        opaque_id = None

        observation = check.not_none(record.event_log_entry.asset_observation)

        for metadata_key in observation.metadata:
            if metadata_key.startswith(OPAQUE_ID_METADATA_KEY_PREFIX):
                opaque_id = metadata_key[len(OPAQUE_ID_METADATA_KEY_PREFIX) :]
                break
        if opaque_id is not None:
            yield (
                AssetMaterializationId(
                    run_id=run_id,
                    asset_key=asset_key,
                    partition=partition,
                    step_key=step_key,
                ),
                opaque_id,
            )


QUERY_HISTORY_TIME_PADDING = timedelta(hours=1)  # deal with desynchronized clocks


@dataclass
class SnowflakeInsightsDefinitions:
    assets: Sequence[AssetsDefinition]
    schedule: ScheduleDefinition


def get_cost_data_for_hour(
    snowflake: "SnowflakeConnection",
    start_hour: datetime,
    end_hour: datetime,
) -> List[Tuple[str, float, str]]:
    """Given a date range, queries the Snowflake query_history table for all queries that were run
    during that time period and returns a mapping from AssetMaterializationId to the cost of the
    query that produced it, as estimated by Snowflake. The cost is in Snowflake credits.
    """
    opaque_ids_sql = rf"""
    regexp_substr_all(query_text, '{OPAQUE_ID_SQL_SIGIL}\\[\\[\\[(.*?)\\]\\]\\]', 1, 1, 'ce', 1)
    """.strip()

    sql = f"""
WITH
warehouse_sizes AS (
    SELECT 'X-Small' AS warehouse_size, 1 AS credits_per_hour UNION ALL
    SELECT 'Small' AS warehouse_size, 2 AS credits_per_hour UNION ALL
    SELECT 'Medium'  AS warehouse_size, 4 AS credits_per_hour UNION ALL
    SELECT 'Large' AS warehouse_size, 8 AS credits_per_hour UNION ALL
    SELECT 'X-Large' AS warehouse_size, 16 AS credits_per_hour UNION ALL
    SELECT '2X-Large' AS warehouse_size, 32 AS credits_per_hour UNION ALL
    SELECT '3X-Large' AS warehouse_size, 64 AS credits_per_hour UNION ALL
    SELECT '4X-Large' AS warehouse_size, 128 AS credits_per_hour UNION ALL
    SELECT 'missing' as warehouse_size, 0 as credits_per_hour
)
SELECT
    {opaque_ids_sql} as opaque_ids,
    qh.execution_time/(1000*60*60)*wh.credits_per_hour AS query_cost,
    query_id
FROM snowflake.account_usage.query_history AS qh
INNER JOIN warehouse_sizes AS wh
    ON ifnull(qh.warehouse_size, 'missing')=wh.warehouse_size
WHERE
    start_time >= '{start_hour - QUERY_HISTORY_TIME_PADDING}'
    AND start_time <= '{end_hour + QUERY_HISTORY_TIME_PADDING}'
HAVING ARRAY_SIZE(opaque_ids) > 0
        """

    with snowflake.get_connection() as conn:
        with conn.cursor() as cursor:
            result = cursor.execute(sql)
            assert result
            results = result.fetchall()

    costs: List[Tuple[str, float, str]] = []

    print(
        f"{len(results) if results else 0} annotated queries returned from snowflake query_history"
    )

    if not results:
        return []

    total = 0
    for result_opaque_ids, result_cost, query_id in results:
        opaque_ids = json.loads(result_opaque_ids)
        # TODO: is this cost splitting logic correct?
        cost = result_cost / len(opaque_ids)
        total += len(opaque_ids)
        for opaque_id in opaque_ids:
            costs.append((opaque_id, float(cost), query_id))

    print(
        f"Reported costs for {len(costs)} of {total} asset materializations found in the"
        " query_history."
    )
    return costs



from dagster import (
    Array,
    BoolSource,
    Field,
    IntSource,
    Map,
    Noneable,
    Permissive,
    ScalarUnion,
    StringSource,
)
from dagster_cloud_cli.core.graphql_client import (
    DEFAULT_BACKOFF_FACTOR,
    DEFAULT_RETRIES,
    DEFAULT_TIMEOUT,
)


def dagster_cloud_api_config():
    return {
        "url": Field(StringSource, is_required=False),
        "agent_token": Field(StringSource, is_required=True),
        "headers": Field(Permissive(), default_value={}),
        "cookies": Field(Permissive(), default_value={}),
        "timeout": Field(Noneable(IntSource), default_value=DEFAULT_TIMEOUT),
        "verify": Field(bool, default_value=True),
        "retries": Field(IntSource, default_value=DEFAULT_RETRIES),
        "method": Field(StringSource, default_value="POST"),
        "backoff_factor": Field(float, default_value=DEFAULT_BACKOFF_FACTOR),
        "proxies": Field(Map(str, str), is_required=False),
        "agent_label": Field(StringSource, is_required=False),
        # Handle requests for a single non-branch deployment
        "deployment": Field(
            ScalarUnion(scalar_type=str, non_scalar_schema=Array(str)), is_required=False
        ),
        # Handle requests for multiple non-branch deployments
        "deployments": Field(Array(StringSource), is_required=False),
        # Handle requests for all branch deployments (can be combined with `deployment`)`
        "branch_deployments": Field(BoolSource, default_value=False, is_required=False),
        "all_serverless_deployments": Field(BoolSource, default_value=False, is_required=False),
    }



from dagster import (
    Array,
    BoolSource,
    Field,
    Noneable,
    Permissive,
    Shape,
    StringSource,
)

SHARED_K8S_CONFIG = {
    "namespace": Field(
        Noneable(StringSource),
        is_required=False,
        description=(
            "The namespace into which to launch Kubernetes resources. Note that any "
            "other required resources (such as the service account) must be "
            "present in this namespace."
        ),
    ),
    "image_pull_policy": Field(
        Noneable(StringSource),
        is_required=False,
        description="Image pull policy to set on launched Pods.",
    ),
    "env_config_maps": Field(
        Noneable(Array(StringSource)),
        is_required=False,
        description=(
            "A list of custom ConfigMapEnvSource names from which to draw "
            "environment variables (using ``envFrom``) for the Job. Default: ``[]``. See:"
            "https://kubernetes.io/docs/tasks/inject-data-application/define-environment-variable-container/#define-an-environment-variable-for-a-container"
        ),
    ),
    "env_secrets": Field(
        Noneable(Array(StringSource)),
        is_required=False,
        description=(
            "A list of custom Secret names from which to draw environment "
            "variables (using ``envFrom``) for the Job. Default: ``[]``. See:"
            "https://kubernetes.io/docs/tasks/inject-data-application/distribute-credentials-secure/#configure-all-key-value-pairs-in-a-secret-as-container-environment-variables"
        ),
    ),
    "service_account_name": Field(
        Noneable(StringSource),
        is_required=False,
        description="The name of the Kubernetes service account under which to run.",
    ),
    "env_vars": Field(
        Noneable(Array(str)),
        is_required=False,
        description=(
            "A list of environment variables to inject into the Job. Each can be "
            "of the form KEY=VALUE or just KEY (in which case the value will be pulled from "
            "the current process). Default: ``[]``. See: "
            "https://kubernetes.io/docs/tasks/inject-data-application/distribute-credentials-secure/#configure-all-key-value-pairs-in-a-secret-as-container-environment-variables"
        ),
    ),
    "volume_mounts": Field(
        Array(
            # Can supply either snake_case or camelCase, but in typeaheads based on the
            # schema we assume snake_case
            Permissive(
                {
                    "name": StringSource,
                    "mount_path": Field(StringSource, is_required=False),
                    "mount_propagation": Field(StringSource, is_required=False),
                    "read_only": Field(BoolSource, is_required=False),
                    "sub_path": Field(StringSource, is_required=False),
                    "sub_path_expr": Field(StringSource, is_required=False),
                }
            )
        ),
        is_required=False,
        default_value=[],
        description=(
            "A list of volume mounts to include in the job's container. Default: ``[]``. See: "
            "https://v1-18.docs.kubernetes.io/docs/reference/generated/kubernetes-api/v1.18/#volumemount-v1-core"
        ),
    ),
    "volumes": Field(
        Array(
            Permissive(
                {
                    "name": str,
                }
            )
        ),
        is_required=False,
        default_value=[],
        description=(
            "A list of volumes to include in the Job's Pod. Default: ``[]``. For the many "
            "possible volume source types that can be included, see: "
            "https://v1-18.docs.kubernetes.io/docs/reference/generated/kubernetes-api/v1.18/#volume-v1-core"
        ),
    ),
    "image_pull_secrets": Field(
        Noneable(Array(Shape({"name": StringSource}))),
        is_required=False,
        description=(
            "Specifies that Kubernetes should get the credentials from "
            "the Secrets named in this list."
        ),
    ),
    "labels": Field(
        dict,
        is_required=False,
        description=(
            "Labels to apply to all created pods. See: "
            "https://kubernetes.io/docs/concepts/overview/working-with-objects/labels"
        ),
    ),
    "resources": Field(
        Noneable(
            {
                "limits": Field(dict, is_required=False),
                "requests": Field(dict, is_required=False),
            }
        ),
        is_required=False,
        description=(
            "Compute resource requirements for the container. See: "
            "https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/"
        ),
    ),
    "scheduler_name": Field(
        Noneable(StringSource),
        is_required=False,
        description=(
            "Use a custom Kubernetes scheduler for launched Pods. See:"
            "https://kubernetes.io/docs/tasks/extend-kubernetes/configure-multiple-schedulers/"
        ),
    ),
    "security_context": Field(
        dict,
        is_required=False,
        description=(
            "Security settings for the container. See:"
            "https://kubernetes.io/docs/tasks/configure-pod-container/security-context/#set-capabilities-for-a-container"
        ),
    ),
}



import copy
import re
import time
from typing import Mapping, Optional

import kubernetes
from dagster._utils.merger import merge_dicts
from dagster_k8s.client import DagsterKubernetesClient
from dagster_k8s.models import k8s_model_from_dict
from kubernetes import client

from dagster_cloud.instance import DagsterCloudAgentInstance

from ..user_code_launcher.utils import (
    deterministic_label_for_location,
    get_human_readable_label,
    unique_resource_name,
)

MANAGED_RESOURCES_LABEL = {"managed_by": "K8sUserCodeLauncher"}
SERVICE_PORT = 4000


def _get_dagster_k8s_labels(
    deployment_name: str, location_name: str, instance: DagsterCloudAgentInstance
) -> Mapping[str, str]:
    return {
        **MANAGED_RESOURCES_LABEL,
        "location_hash": deterministic_label_for_location(deployment_name, location_name),
        "location_name": get_k8s_human_readable_label(location_name),
        "deployment_name": get_k8s_human_readable_label(deployment_name),
        "agent_id": instance.instance_uuid,
    }


def _sanitize_k8s_resource_name(name):
    filtered_name = re.sub("[^a-z0-9-]", "", name.lower())

    # ensure it doesn't start with a non-alpha character
    while filtered_name and re.match("[^a-z].*", filtered_name):
        filtered_name = filtered_name[1:]

    filtered_name = filtered_name.strip("-")

    # always return something that starts with a letter in the unlikely event that everything is
    # filtered out (doesn't have to be unique)
    return filtered_name or "k8s"


def unique_k8s_resource_name(deployment_name, location_name):
    """https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#dns-label-names.

    K8s resource names are restricted, so we must sanitize the location name to not include disallowed characters.
    """
    return unique_resource_name(
        deployment_name, location_name, length_limit=63, sanitize_fn=_sanitize_k8s_resource_name
    )


def get_k8s_human_readable_label(name):
    """https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set.

    K8s label values are restricted, so we must sanitize the location name to not include disallowed characters.
    These are purely to help humans debug, so they don't need to be unique.
    """
    return get_human_readable_label(
        name,
        length_limit=63,
        sanitize_fn=lambda name: (
            re.sub("[^a-zA-Z0-9-_.]", "", name).strip("-").strip("_").strip(".")
        ),
    )


def construct_code_location_service(
    deployment_name, location_name, service_name, container_context, instance
):
    labels = container_context.labels

    return client.V1Service(
        metadata=client.V1ObjectMeta(
            name=service_name,
            labels={
                **labels,
                **_get_dagster_k8s_labels(deployment_name, location_name, instance),
            },
        ),
        spec=client.V1ServiceSpec(
            selector={"user-deployment": service_name},
            ports=[client.V1ServicePort(name="grpc", protocol="TCP", port=SERVICE_PORT)],
        ),
    )


def construct_code_location_deployment(
    instance,
    deployment_name,
    location_name,
    k8s_deployment_name,
    metadata,
    container_context,
    args,
):
    pull_policy = container_context.image_pull_policy
    env_config_maps = container_context.env_config_maps
    env_secrets = container_context.env_secrets
    service_account_name = container_context.service_account_name
    image_pull_secrets = container_context.image_pull_secrets
    volume_mounts = container_context.volume_mounts

    volumes = container_context.volumes
    resources = container_context.resources

    scheduler_name = container_context.scheduler_name
    security_context = container_context.security_context

    env = merge_dicts(
        metadata.get_grpc_server_env(
            SERVICE_PORT, location_name, instance.ref_for_deployment(deployment_name)
        ),
        container_context.get_environment_dict(),
    )

    user_defined_config = container_context.server_k8s_config

    container_config = copy.deepcopy(user_defined_config.container_config)

    container_config["args"] = args

    if pull_policy:
        container_config["image_pull_policy"] = pull_policy

    user_defined_env_vars = container_config.pop("env", [])
    user_defined_env_from = container_config.pop("env_from", [])
    user_defined_volume_mounts = container_config.pop("volume_mounts", [])
    user_defined_resources = container_config.pop("resources", {})
    user_defined_security_context = container_config.pop("security_context", None)

    container_name = container_config.get("name", "dagster")

    container_config = {
        **container_config,
        "name": container_name,
        "image": metadata.image,
        "env": [{"name": key, "value": value} for key, value in env.items()]
        + user_defined_env_vars,
        "env_from": (
            [{"config_map_ref": {"name": config_map}} for config_map in env_config_maps]
            + [{"secret_ref": {"name": secret_name}} for secret_name in env_secrets]
            + user_defined_env_from
        ),
        "volume_mounts": volume_mounts + user_defined_volume_mounts,
        "resources": user_defined_resources or resources,
        "security_context": user_defined_security_context or security_context,
    }

    pod_spec_config = copy.deepcopy(user_defined_config.pod_spec_config)

    user_defined_image_pull_secrets = pod_spec_config.pop("image_pull_secrets", [])
    user_defined_service_account_name = pod_spec_config.pop("service_account_name", None)
    user_defined_containers = pod_spec_config.pop("containers", [])
    user_defined_volumes = pod_spec_config.pop("volumes", [])
    user_defined_scheduler_name = pod_spec_config.pop("scheduler_name", None)

    pod_spec_config = {
        **pod_spec_config,
        "image_pull_secrets": [{"name": x["name"]} for x in image_pull_secrets]
        + user_defined_image_pull_secrets,
        "service_account_name": user_defined_service_account_name or service_account_name,
        "containers": [container_config] + user_defined_containers,
        "volumes": volumes + user_defined_volumes,
        "scheduler_name": user_defined_scheduler_name or scheduler_name,
    }

    pod_template_spec_metadata = copy.deepcopy(user_defined_config.pod_template_spec_metadata)
    user_defined_pod_template_labels = pod_template_spec_metadata.pop("labels", {})

    deployment_dict = {
        "metadata": {
            "name": k8s_deployment_name,
            "labels": {
                **container_context.labels,
                **_get_dagster_k8s_labels(deployment_name, location_name, instance),
            },
        },
        "spec": {  # DeploymentSpec
            "selector": {"match_labels": {"user-deployment": k8s_deployment_name}},
            "template": {  # PodTemplateSpec
                "metadata": {
                    **pod_template_spec_metadata,
                    "labels": {
                        "user-deployment": k8s_deployment_name,
                        **container_context.labels,
                        **user_defined_pod_template_labels,
                        **_get_dagster_k8s_labels(deployment_name, location_name, instance),
                    },
                },
                "spec": pod_spec_config,
            },
        },
    }

    return k8s_model_from_dict(
        kubernetes.client.V1Deployment,
        deployment_dict,
    )


def get_container_waiting_reason(pod) -> Optional[str]:
    if (not pod.status.container_statuses) or len(pod.status.container_statuses) == 0:
        return None

    container_waiting_state = pod.status.container_statuses[0].state.waiting
    if not container_waiting_state:
        return None

    return container_waiting_state.reason


def get_deployment_failure_debug_info(
    k8s_deployment_name, namespace, core_api_client, pod_list, logger
):
    if not pod_list:
        return (
            "For more information about the failure, run `kubectl describe deployment"
            f" {k8s_deployment_name}` in your cluster."
        )

    pod = pod_list[0]
    pod_name = pod.metadata.name

    kubectl_prompt = (
        f"For more information about the failure, run `kubectl describe pod {pod_name}`"
        f" or `kubectl describe deployment {k8s_deployment_name}` in your cluster."
    )
    pod_debug_info = ""
    try:
        api_client = DagsterKubernetesClient.production_client(core_api_override=core_api_client)
        pod_debug_info = api_client.get_pod_debug_info(pod_name, namespace)
    except Exception:
        logger.exception(f"Error trying to get debug information for failed k8s pod {pod_name}")

    return f"{pod_debug_info}\n\n{kubectl_prompt}" if pod_debug_info else kubectl_prompt


def wait_for_deployment_complete(
    k8s_deployment_name,
    namespace,
    logger,
    location_name,
    metadata,
    timeout,
    image_pull_grace_period,
    core_api,
):
    """Translated from
    https://github.com/kubernetes/kubectl/blob/ac49920c0ccb0dd0899d5300fc43713ee2dfcdc9/pkg/polymorphichelpers/rollout_status.go#L75-L91.
    """
    api = client.AppsV1Api(client.ApiClient())

    start = time.time()

    pod_list = []

    while True:
        time.sleep(2)

        time_elapsed = time.time() - start

        if time_elapsed >= timeout:
            timeout_message: str = f"Timed out waiting for deployment {k8s_deployment_name}."
            debug_info = get_deployment_failure_debug_info(
                k8s_deployment_name, namespace, core_api, pod_list, logger
            )
            if debug_info:
                timeout_message = timeout_message + "\n\n" + debug_info

            raise Exception(timeout_message)

        deployment = api.read_namespaced_deployment(k8s_deployment_name, namespace)
        status = deployment.status
        spec = deployment.spec

        logger.debug(
            f"[updated_replicas:{status.updated_replicas},replicas:{status.replicas},available_replicas:{status.available_replicas},observed_generation:{status.observed_generation}]"
            " waiting..."
        )
        logger.debug(f"Status: {status}, spec: {spec}")

        if (
            status.updated_replicas == spec.replicas  # new replicas have been updated
            and status.replicas == status.updated_replicas  # no old replicas pending termination
            and status.available_replicas == status.updated_replicas  # updated replicas available
            and status.observed_generation >= deployment.metadata.generation  # new spec observed
        ):
            return True

        pod_list = core_api.list_namespaced_pod(
            namespace, label_selector=f"user-deployment={k8s_deployment_name}"
        ).items

        if time_elapsed >= image_pull_grace_period:
            for pod in pod_list:
                waiting_reason = get_container_waiting_reason(pod)
                if (
                    waiting_reason == "ImagePullBackOff"
                    or waiting_reason == "ErrImageNeverPull"
                    or waiting_reason == "CreateContainerConfigError"
                ):
                    error_message = f"Error creating deployment for {k8s_deployment_name}."
                    debug_info = get_deployment_failure_debug_info(
                        k8s_deployment_name, namespace, core_api, pod_list, logger
                    )
                    if debug_info:
                        error_message = error_message + "\n" + debug_info

                    raise Exception(error_message)



from abc import abstractproperty
from enum import Enum
from typing import Dict

from dagster._core.definitions.metadata import FloatMetadataValue, MetadataValue
from pydantic import BaseModel


class AnomalyDetectionModelVersion(Enum):
    FRESHNESS_BETA = "FRESHNESS_BETA"


### INTERNAL MODEL PARAMETER SETS ###


class InternalModelParams(BaseModel):
    @abstractproperty
    def model_version(self) -> AnomalyDetectionModelVersion:
        raise NotImplementedError("Subclasses must implement this method")


class InternalBetaFreshnessAnomalyDetectionParams(InternalModelParams):
    sensitivity: float
    asset_key_user_string: str

    @property
    def model_version(self) -> AnomalyDetectionModelVersion:
        return AnomalyDetectionModelVersion.FRESHNESS_BETA


### USER FACING MODEL PARAMETER SETS ###


class AnomalyDetectionModelParams(BaseModel):
    @abstractproperty
    def model_version(self) -> AnomalyDetectionModelVersion:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractproperty
    def as_metadata(self) -> Dict[str, MetadataValue]:
        raise NotImplementedError("Subclasses must implement this method")


class BetaFreshnessAnomalyDetectionParams(AnomalyDetectionModelParams):
    sensitivity: float

    @property
    def model_version(self) -> AnomalyDetectionModelVersion:
        return AnomalyDetectionModelVersion.FRESHNESS_BETA

    @property
    def as_metadata(self) -> Dict[str, MetadataValue]:
        return {
            "sensitivity": FloatMetadataValue(self.sensitivity),
        }



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    Optional,
    Union,
)

import dagster._check as check
from dagster import (
    AssetCheckResult,
    AssetExecutionContext,
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    MetadataValue,
    OpExecutionContext,
    Output,
)
from dagster._core.definitions.metadata import MetadataMapping
from google.cloud import bigquery

from ..insights_utils import extract_asset_info_from_event
from .bigquery_utils import build_bigquery_cost_metadata, marker_asset_key_for_job

if TYPE_CHECKING:
    from dagster_dbt import DbtCliInvocation


@dataclass
class BigQueryCostInfo:
    asset_key: AssetKey
    partition: Optional[str]
    job_id: Optional[str]
    slots_ms: int
    bytes_billed: int

    @property
    def asset_partition_key(self) -> str:
        return (
            f"{self.asset_key.to_string()}:{self.partition}"
            if self.partition
            else self.asset_key.to_string()
        )


def _extract_metadata_value(value: Optional[MetadataValue], default_value: Any = None) -> Any:
    return value.value if value else default_value


def _extract_bigquery_info_from_metadata(metadata: MetadataMapping):
    job_id = _extract_metadata_value(metadata.get("job_id"))
    bytes_billed = _extract_metadata_value(metadata.get("bytes_billed"), 0)
    slots_ms = _extract_metadata_value(metadata.get("slots_ms"), 0)
    return job_id, bytes_billed, slots_ms


def _asset_partition_key(asset_key: AssetKey, partition_key: Optional[str]) -> str:
    return f"{asset_key.to_string()}:{partition_key}" if partition_key else asset_key.to_string()


def dbt_with_bigquery_insights(
    context: Union[OpExecutionContext, AssetExecutionContext],
    dbt_cli_invocation: "DbtCliInvocation",
    dagster_events: Optional[
        Iterable[Union[Output, AssetMaterialization, AssetObservation, AssetCheckResult]]
    ] = None,
    record_observation_usage: bool = True,
    explicitly_query_information_schema: bool = False,
    bigquery_client: Optional[bigquery.Client] = None,
    bigquery_location: Optional[str] = None,
) -> Iterator[Union[Output, AssetMaterialization, AssetObservation, AssetCheckResult]]:
    """Wraps a dagster-dbt invocation to associate each BigQuery query with the produced
    asset materializations. This allows the cost of each query to be associated with the asset
    materialization that it produced.

    If called in the context of an op (rather than an asset), filters out any Output events
    which do not correspond with any output of the op.

    Args:
        context (AssetExecutionContext): The context of the asset that is being materialized.
        dbt_cli_invocation (DbtCliInvocation): The invocation of the dbt CLI to wrap.
        dagster_events (Optional[Iterable[Union[Output, AssetObservation, AssetCheckResult]]]):
            The events that were produced by the dbt CLI invocation. If not provided, it is assumed
            that the dbt CLI invocation has not yet been run, and it will be run and the events
            will be streamed.
        record_observation_usage (bool): If True, associates the usage associated with
            asset observations with that asset. Default is True.

    **Example:**

    .. code-block:: python

        @dbt_assets(manifest=DBT_MANIFEST_PATH)
        def jaffle_shop_dbt_assets(
            context: AssetExecutionContext,
            dbt: DbtCliResource,
        ):
            dbt_cli_invocation = dbt.cli(["build"], context=context)
            yield from dbt_with_bigquery_insights(context, dbt_cli_invocation)
    """
    if explicitly_query_information_schema:
        check.inst_param(bigquery_client, "bigquery_client", bigquery.Client)
        check.str_param(bigquery_location, "bigquery_location")

    if dagster_events is None:
        dagster_events = dbt_cli_invocation.stream()

    asset_info_by_job_id = {}
    cost_by_asset = defaultdict(list)
    for dagster_event in dagster_events:
        if isinstance(dagster_event, (AssetMaterialization, AssetObservation, Output)):
            asset_key, partition = extract_asset_info_from_event(
                context, dagster_event, record_observation_usage
            )
            if not asset_key:
                asset_key = marker_asset_key_for_job(context.job_def)
            job_id, bytes_billed, slots_ms = _extract_bigquery_info_from_metadata(
                dagster_event.metadata
            )
            if bytes_billed or slots_ms:
                asset_info_by_job_id[job_id] = (asset_key, partition)
                cost_info = BigQueryCostInfo(asset_key, partition, job_id, slots_ms, bytes_billed)
                cost_by_asset[cost_info.asset_partition_key].append(cost_info)

        yield dagster_event

    if explicitly_query_information_schema and bigquery_client:
        marker_asset_key = marker_asset_key_for_job(context.job_def)
        run_results_json = dbt_cli_invocation.get_artifact("run_results.json")
        invocation_id = run_results_json["metadata"]["invocation_id"]
        assert bigquery_client
        try:
            cost_query = f"""
                SELECT job_id, SUM(total_bytes_billed) AS bytes_billed, SUM(total_slot_ms) AS slots_ms
                FROM `{bigquery_location}`.INFORMATION_SCHEMA.JOBS
                WHERE query like '%{invocation_id}%'
                GROUP BY job_id
            """
            context.log.info(f"Querying INFORMATION_SCHEMA.JOBS for bytes billed: {cost_query}")
            query_result = bigquery_client.query(cost_query)

            # overwrite cost_by_asset that is computed from the metadata
            cost_by_asset = defaultdict(list)
            for row in query_result:
                asset_key, partition = asset_info_by_job_id.get(
                    row.job_id, (marker_asset_key, None)
                )
                if row.bytes_billed or row.slots_ms:
                    cost_info = BigQueryCostInfo(
                        asset_key, partition, row.job_id, row.bytes_billed, row.slots_ms
                    )
                    cost_by_asset[cost_info.asset_partition_key].append(cost_info)
        except:
            context.log.exception("Could not query information_schema.jobs for bytes billed")

    for cost_info_list in cost_by_asset.values():
        bytes_billed = sum(item.bytes_billed for item in cost_info_list)
        slots_ms = sum(item.slots_ms for item in cost_info_list)
        job_ids = [item.job_id for item in cost_info_list]
        asset_key = cost_info_list[0].asset_key
        partition = cost_info_list[0].partition
        context.log_event(
            AssetObservation(
                asset_key=asset_key,
                partition=partition,
                metadata=build_bigquery_cost_metadata(job_ids, bytes_billed, slots_ms),
            )
        )



import json
import logging
import time
from typing import List, Optional

import boto3
import botocore
import dagster._check as check
from botocore.config import Config
from botocore.exceptions import ClientError
from dagster._utils.cached_method import cached_method
from dagster_aws.ecs.tasks import DagsterEcsTaskDefinitionConfig
from dagster_aws.ecs.utils import task_definitions_match

from dagster_cloud.workspace.ecs.service import Service

DEFAULT_ECS_TIMEOUT = 300
DEFAULT_ECS_GRACE_PERIOD = 30

STOPPED_TASK_GRACE_PERIOD = 30


ECS_EXEC_LINUX_PARAMETERS = {
    "capabilities": {"add": ["SYS_PTRACE"]},
    "initProcessEnabled": True,
}

config = Config(retries={"max_attempts": 50, "mode": "standard"})


def get_debug_ecs_prompt(cluster: str, task_arn: str) -> str:
    return (
        "For more information about the failure, check the ECS console for logs for task"
        f" {task_arn} in cluster {cluster}."
    )


class EcsServiceError(Exception):
    def __init__(self, cluster, task_arn, stopped_reason, logs, show_debug_prompt: bool):
        log_str = "Task logs:\n" + "\n".join(logs) if logs else "No task logs."

        message = (
            f"ECS service failed because task {task_arn} failed: {stopped_reason}\n\n{log_str}"
        )

        if show_debug_prompt:
            message = message + f"\n\n{get_debug_ecs_prompt(cluster, task_arn)}"
        super().__init__(message)


class Client:
    def __init__(
        self,
        cluster_name: str,
        service_discovery_namespace_id: str,
        log_group: str,
        subnet_ids: Optional[List[str]] = None,
        security_group_ids: Optional[List[str]] = None,
        ecs_client=None,
        timeout: int = DEFAULT_ECS_TIMEOUT,
        grace_period: int = DEFAULT_ECS_GRACE_PERIOD,
        launch_type: str = "FARGATE",
        show_debug_cluster_info: bool = True,
    ):
        self.ecs = ecs_client if ecs_client else boto3.client("ecs", config=config)
        self.logs = boto3.client("logs", config=config)
        self.service_discovery = boto3.client("servicediscovery", config=config)
        self.tags_client = boto3.client("resourcegroupstaggingapi", config=config)
        self._use_legacy_tag_filtering = False

        self.cluster_name = cluster_name.split("/")[-1]
        self.subnet_ids = check.opt_list_param(subnet_ids, "subnet_ids")
        self.security_group_ids = check.opt_list_param(security_group_ids, "security_group_ids")
        self.service_discovery_namespace_id = check.str_param(
            service_discovery_namespace_id, "service_discovery_namespace_id"
        )
        self.log_group = check.str_param(log_group, "log_group")

        self.show_debug_cluster_info = show_debug_cluster_info

        self.timeout = check.int_param(timeout, "timeout")
        self.grace_period = check.int_param(grace_period, "grace_period")
        self.launch_type = check.str_param(launch_type, "launch_type")
        self._namespace: Optional[str] = None

    @property
    def ec2(self):
        # Reconstruct the resource on each call because it's not threadsafe
        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/resources.html#multithreading-or-multiprocessing-with-resources
        return boto3.resource("ec2", config=config)

    @property
    def namespace(self):
        if not self._namespace:
            self._namespace = (
                self.service_discovery.get_namespace(
                    Id=self.service_discovery_namespace_id,
                )
                .get("Namespace")
                .get("Name")
            )
        return self._namespace

    @property
    def taggable(self):
        settings = self.ecs.list_account_settings(
            name="serviceLongArnFormat",
            effectiveSettings=True,
        )
        return settings["settings"][0]["value"] == "enabled"

    @property
    @cached_method
    def network_configuration(self):
        network_configuration = {
            "awsvpcConfiguration": {
                "subnets": self.subnet_ids,
                **(
                    {"assignPublicIp": self._assign_public_ip()}
                    if self.launch_type == "FARGATE"
                    else {}
                ),
            },
        }

        if self.security_group_ids:
            network_configuration["awsvpcConfiguration"]["securityGroups"] = self.security_group_ids

        return network_configuration

    def register_task_definition(
        self,
        family,
        image,
        command,
        execution_role_arn,
        container_name=None,
        task_role_arn=None,
        env=None,
        secrets=None,
        sidecars=None,
        logger=None,
        cpu=None,
        memory=None,
        ephemeral_storage=None,
        repository_credentials=None,
        runtime_platform=None,
        mount_points=None,
        volumes=None,
        linux_parameters=None,
    ):
        container_name = container_name or family

        logger = logger or logging.getLogger("dagster_cloud.EcsClient")

        env = env or {}
        environment = [{"name": key, "value": value} for key, value in env.items()]

        secrets = (
            [{"name": key, "valueFrom": value} for key, value in secrets.items()] if secrets else []
        )

        task_definition_config = DagsterEcsTaskDefinitionConfig(
            family=family,
            image=image,
            container_name=container_name,
            command=command,
            log_configuration={
                "logDriver": "awslogs",
                "options": {
                    "awslogs-group": self.log_group,
                    "awslogs-region": self.ecs.meta.region_name,
                    "awslogs-stream-prefix": family,
                },
            },
            secrets=secrets,
            environment=environment,
            execution_role_arn=execution_role_arn,
            task_role_arn=task_role_arn,
            sidecars=sidecars,
            requires_compatibilities=[self.launch_type],
            cpu=cpu,
            memory=memory,
            ephemeral_storage=ephemeral_storage,
            repository_credentials=repository_credentials,
            runtime_platform=runtime_platform,
            mount_points=mount_points,
            volumes=volumes,
            linux_parameters=linux_parameters,
        )

        try:
            existing_task_definition = self.ecs.describe_task_definition(taskDefinition=family)[
                "taskDefinition"
            ]
        except ClientError:
            logger.info("No existing task definition")
            # task definition does not exist, do not reuse
            existing_task_definition = None

        if not existing_task_definition or not task_definitions_match(
            task_definition_config,
            existing_task_definition,
            container_name=container_name,
        ):
            task_definition_arn = (
                self.ecs.register_task_definition(**task_definition_config.task_definition_dict())
                .get("taskDefinition")
                .get("taskDefinitionArn")
            )
            logger.info(f"Created new task definition {task_definition_arn}")
        else:
            task_definition_arn = existing_task_definition["taskDefinitionArn"]
            logger.info(f"Re-using existing task definition {task_definition_arn}")

        return task_definition_arn

    def create_service(
        self,
        name,
        image,
        command,
        execution_role_arn,
        family=None,
        container_name=None,
        task_role_arn=None,
        env=None,
        tags=None,
        register_service_discovery=True,
        secrets=None,
        sidecars=None,
        logger=None,
        cpu=None,
        memory=None,
        ephemeral_storage=None,
        replica_count=None,
        repository_credentials=None,
        allow_ecs_exec=False,
        runtime_platform=None,
        mount_points=None,
        volumes=None,
    ):
        logger = logger or logging.getLogger("dagster_cloud.EcsClient")

        service_name = name
        family = family or name

        logger.info(f"Checking if task definition {family} for {name} can be re-used...")

        task_definition_arn = self.register_task_definition(
            family=family,
            image=image,
            container_name=container_name,
            execution_role_arn=execution_role_arn,
            task_role_arn=task_role_arn,
            command=command,
            env=env or {},
            secrets=secrets,
            sidecars=sidecars,
            logger=logger,
            cpu=cpu,
            memory=memory,
            ephemeral_storage=ephemeral_storage,
            repository_credentials=repository_credentials,
            runtime_platform=runtime_platform,
            mount_points=mount_points,
            volumes=volumes,
            linux_parameters=ECS_EXEC_LINUX_PARAMETERS if allow_ecs_exec else None,
        )

        service_registry_arn = None
        # Configure service discovery
        if register_service_discovery:
            logger.info(f"Creating service registry for {service_name}...")
            service_registry_arn = self._create_service_registry(
                service_name=service_name,
                tags=tags,
            )

        logger.info(f"Creating ECS service {service_name}...")

        # Create the service
        return self._create_service(
            service_name=service_name,
            service_registry_arn=service_registry_arn,
            task_definition_arn=task_definition_arn,
            replica_count=replica_count,
            tags=tags,
            allow_ecs_exec=allow_ecs_exec,
        )

    def delete_service(
        self,
        service,
        logger=None,
    ):
        logger = logger or logging.getLogger("dagster_cloud.EcsClient")

        # Reduce running tasks to 0
        try:
            self.ecs.update_service(
                cluster=self.cluster_name,
                service=service.name,
                desiredCount=0,
            )
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] in [
                "ServiceNotFoundException",
                "ServiceNotActiveException",
            ]:
                logger.warning(
                    f"Service {self.cluster_name}/{service.name} is not found or inactive: {error}"
                )
                return
            raise

        # Delete the ECS service
        self.ecs.delete_service(
            cluster=self.cluster_name,
            service=service.name,
            force=True,
        )
        # get service discovery id
        service_discovery_id = self._get_service_discovery_id(
            service.hostname,
        )
        if service_discovery_id:
            # Unregister dangling ecs tasks from service discovery
            instances_paginator = self.service_discovery.get_paginator("list_instances")
            instances = instances_paginator.paginate(
                ServiceId=service_discovery_id,
            ).build_full_result()["Instances"]
            deregister_operation_ids = []
            for instance in instances:
                resp = self.service_discovery.deregister_instance(
                    ServiceId=service_discovery_id, InstanceId=instance["Id"]
                )
                deregister_operation_ids.append(resp["OperationId"])

            # wait for instances to complete deregistering
            for operation_id in deregister_operation_ids:
                status = ""
                while status != "SUCCESS":
                    status = self.service_discovery.get_operation(OperationId=operation_id)[
                        "Operation"
                    ]["Status"]
                    if status == "FAIL":
                        raise Exception("deregister operation failed")
                    time.sleep(2)

            # delete service discovery
            self.service_discovery.delete_service(
                Id=service_discovery_id,
            )

    def list_services(self, tags=None, logger=None):
        logger = logger or logging.getLogger("dagster_cloud.EcsClient")

        services = []

        if tags:
            if not self._use_legacy_tag_filtering:
                try:
                    # obtain services present in the cluster
                    ecs_services_paginator = self.ecs.get_paginator("list_services")
                    actual_services = []
                    for page in ecs_services_paginator.paginate(cluster=self.cluster_name):
                        actual_services.extend(page["serviceArns"])

                    # obtain services with the specified tags
                    tag_paginator = self.tags_client.get_paginator("get_resources")
                    for page in tag_paginator.paginate(
                        TagFilters=[
                            {
                                "Key": key,
                                "Values": [value],
                            }
                            for key, value in tags.items()
                        ],
                        ResourceTypeFilters=["ecs:service"],
                    ):
                        for resource in page["ResourceTagMappingList"]:
                            resource_arn = resource["ResourceARN"]
                            # note: the Resource Group Tagging API may return resources corresponding to services
                            # that are no longer present in the cluster, so we filter out those that are not present
                            if resource_arn in actual_services:
                                services.append(Service(client=self, arn=resource_arn))

                except botocore.exceptions.ClientError as error:
                    if error.response["Error"]["Code"] == "AccessDeniedException":
                        self._use_legacy_tag_filtering = True
                        logger.warning(
                            "dagster-cloud 1.4.0 includes performance enhancements that rely on the"
                            " Resource"
                            " Group Tagging API. To take advantage, add 'tags:GetResources'"
                            " permissions to"
                            " your"
                            " task role."
                            " https://docs.aws.amazon.com/resourcegroupstagging/latest/APIReference/overview.html"
                        )
                        services = self.list_services(tags)
                    else:
                        raise
            else:
                filtered_services = [
                    service
                    for service in self.list_services()
                    if tags.items() <= service.tags.items()
                ]
                return filtered_services

        else:
            service_paginator = self.ecs.get_paginator("list_services")
            for page in service_paginator.paginate(cluster=self.cluster_name):
                for service_arn in page.get("serviceArns"):
                    service = Service(client=self, arn=service_arn)
                    services.append(service)

        return services

    def _run_task(self, name, image, command, execution_role_arn):
        """This is no longer used except as an integration test util
        for testing various network configurations.
        """
        task_definition_arn = self.register_task_definition(
            family=name,
            image=image,
            command=command,
            execution_role_arn=execution_role_arn,
        )

        task_arn = (
            self.ecs.run_task(
                taskDefinition=task_definition_arn,
                cluster=self.cluster_name,
                launchType=self.launch_type,
                networkConfiguration=self.network_configuration,
            )
            .get("tasks", [{}])[0]
            .get("taskArn")
        )

        self.ecs.get_waiter("tasks_stopped").wait(
            cluster=self.cluster_name,
            tasks=[task_arn],
            WaiterConfig={"Delay": 1, "MaxAttempts": self.timeout},
        )

        exit_code = (
            self.ecs.describe_tasks(
                cluster=self.cluster_name,
                tasks=[task_arn],
            )
            .get("tasks", [{}])[0]
            .get("containers", [{}])[0]
            .get("exitCode")
        )

        if exit_code:
            raise Exception(self.get_task_logs(task_arn))

        return True

    def _create_service_registry(self, service_name, tags=None):
        if not tags:
            tags = {}

        # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/create-service-discovery.html
        service_registry_arn = (
            self.service_discovery.create_service(
                Name=service_name,
                NamespaceId=self.service_discovery_namespace_id,
                DnsConfig={
                    "DnsRecords": [
                        {"Type": "A", "TTL": 60},
                    ]
                },
                Tags=[{"Key": key, "Value": value} for key, value in tags.items()],
            )
            .get("Service", {})
            .get("Arn")
        )
        return service_registry_arn

    def _create_service(
        self,
        service_name,
        task_definition_arn,
        service_registry_arn,
        replica_count=None,
        tags=None,
        allow_ecs_exec=False,
    ):
        params = dict(
            cluster=self.cluster_name,
            serviceName=service_name,
            taskDefinition=task_definition_arn,
            launchType=self.launch_type,
            desiredCount=replica_count or 1,
            enableExecuteCommand=allow_ecs_exec,
            propagateTags="SERVICE",
        )
        params["networkConfiguration"] = self.network_configuration

        if service_registry_arn:
            params["serviceRegistries"] = [{"registryArn": service_registry_arn}]

        if tags and self.taggable:
            params["tags"] = [
                {"key": key, **({"value": value} if value is not None else {})}
                for key, value in tags.items()
            ]

        arn = self.ecs.create_service(**params).get("service").get("serviceArn")

        return Service(client=self, arn=arn)

    def wait_for_new_service(self, service, container_name, logger=None) -> dict:
        logger = logger or logging.getLogger("dagster_cloud.EcsClient")
        service_name = service.name
        start_time = time.time()
        while True:
            response = self.ecs.describe_services(
                cluster=self.cluster_name,
                services=[service_name],
            )
            if response.get("services"):
                running_tasks = self.check_service_has_running_tasks(
                    service_name, container_name, logger=logger
                )
                return running_tasks[0]

            failures = response.get("failures")

            if not failures:
                # This might not be a possible state; it's unclear if the ECS API can return empty lists for both
                # "failures" and "services":
                # https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_DescribeServices.html
                raise Exception(
                    "ECS DescribeServices API returned an empty response for service"
                    f" {service.arn}."
                )

            # Even if we fail, check a few more times in case it's just the ECS API
            # being eventually consistent
            if time.time() - start_time >= self.grace_period:
                raise Exception(
                    f"ECS DescribeServices API returned failures: {json.dumps(failures)}"
                )

            time.sleep(10)

    def _raise_failed_task(self, task, container_name, logger):
        task_arn = task["taskArn"]
        stopped_reason = task["stoppedReason"]

        logs = None

        try:
            logs = self.get_task_logs(task_arn, container_name=container_name)
        except:
            logger.exception(f"Error trying to get logs for failed task {task_arn}")

        raise EcsServiceError(
            cluster=self.cluster_name,
            task_arn=task_arn,
            stopped_reason=stopped_reason,
            logs=logs,
            show_debug_prompt=self.show_debug_cluster_info,
        )

    def assert_task_not_stopped(self, task_arn, container_name, logger=None):
        logger = logger or logging.getLogger("dagster_cloud.EcsClient")

        task = self.ecs.describe_tasks(cluster=self.cluster_name, tasks=[task_arn]).get("tasks")[0]
        if task.get("lastStatus") == "STOPPED":
            self._raise_failed_task(task, container_name, logger)

    def check_service_has_running_tasks(
        self, service_name, container_name, logger=None
    ) -> List[dict]:
        # return the ARN of the task if it starts
        logger = logger or logging.getLogger("dagster_cloud.EcsClient")
        start_time = time.time()

        tasks_to_track = None

        while start_time + self.timeout > time.time():
            # Check for running tasks to track
            if not tasks_to_track:
                services = self.ecs.describe_services(
                    cluster=self.cluster_name,
                    services=[service_name],
                )
                if not services:
                    raise Exception(
                        f"Service description not found for {self.cluster_name}/{service_name}"
                    )

                service = services.get("services")[0]
                desired_count = service.get("desiredCount")
                running_count = service.get("runningCount")

                # If the service has reached the desired count, we can start tracking the tasks
                if desired_count == running_count:
                    running_tasks = self.ecs.list_tasks(
                        cluster=self.cluster_name,
                        serviceName=service_name,
                        desiredStatus="RUNNING",
                    ).get("taskArns")

                    if not running_tasks:
                        raise Exception(
                            f"Unexpected error obtaining tasks for {service_name} in {self.cluster_name}"
                        )

                    tasks_to_track = running_tasks
                elif time.time() > start_time + STOPPED_TASK_GRACE_PERIOD:
                    # If there are still no running_tasks tasks after a certain grace period, check for stopped tasks
                    stopped_tasks = self._check_for_stopped_tasks(service_name)
                    if stopped_tasks:
                        self._raise_failed_task(stopped_tasks[0], container_name, logger)

            if tasks_to_track:
                tasks = self.ecs.describe_tasks(
                    cluster=self.cluster_name, tasks=tasks_to_track
                ).get("tasks")

                # Wait for all tasks essential containers to be running
                all_tasks_running = len(tasks) > 0
                for task in tasks:
                    if task.get("lastStatus") == "RUNNING":
                        if not self._check_all_essential_containers_are_running(task):
                            all_tasks_running = False
                    elif task.get("lastStatus") == "STOPPED":
                        self._raise_failed_task(task, container_name, logger)

                if all_tasks_running:
                    return tasks_to_track

            time.sleep(20)

        # Fetch the service event logs to try to get some clue about why the service never spun
        # up any tasks
        service_events_str = ""
        try:
            response = self.ecs.describe_services(
                cluster=self.cluster_name,
                services=[service_name],
            )
            if response.get("services"):
                service = response["services"][0]
                service_events = [event.get("message") for event in service.get("events")]
                service_events_str = "Service events:\n" + "\n".join(service_events)
        except:
            logger.exception(f"Error trying to get service event logs from service {service_name}")

        raise Exception(
            f"Timed out waiting for a running task for service: {service_name}."
            f" {service_events_str}"
        )

    def _check_for_stopped_tasks(self, service_name):
        stopped = self.ecs.list_tasks(
            cluster=self.cluster_name,
            serviceName=service_name,
            desiredStatus="STOPPED",
        ).get("taskArns")
        if stopped:
            stopped_tasks = self.ecs.describe_tasks(cluster=self.cluster_name, tasks=stopped).get(
                "tasks"
            )

            stopped_tasks = sorted(
                stopped_tasks,
                key=lambda task: task["createdAt"].timestamp(),
                reverse=True,
            )
            return stopped_tasks
        return []

    def _check_all_essential_containers_are_running(self, task) -> bool:
        task_definition = self.ecs.describe_task_definition(
            taskDefinition=task.get("taskDefinitionArn"),
        ).get("taskDefinition")

        essential_containers = {
            container["name"]
            for container in task_definition["containerDefinitions"]
            if container["essential"]
        }

        # Just because the task is RUNNING doesn't mean everything has started up correctly -
        # sometimes it briefly thinks it is RUNNING even though individual containers have STOPPED.
        # Wait for all essential containers to be running_tasks too before declaring victory.
        return all(
            container["name"] not in essential_containers or container["lastStatus"] == "RUNNING"
            for container in task["containers"]
        )

    def _get_service_discovery_id(self, hostname):
        service_name = hostname.split("." + self.namespace)[0]

        paginator = self.service_discovery.get_paginator("list_services")
        for page in paginator.paginate(
            Filters=[
                {
                    "Name": "NAMESPACE_ID",
                    "Values": [
                        self.service_discovery_namespace_id,
                    ],
                    "Condition": "EQ",
                },
            ],
        ):
            for service in page["Services"]:
                if service["Name"] == service_name:
                    return service["Id"]

    def _assign_public_ip(self):
        # https://docs.aws.amazon.com/AmazonECS/latest/userguide/fargate-task-networking.html
        # Assign a public IP if any of the subnets are public
        route_tables = self.ec2.route_tables.filter(
            Filters=[
                {"Name": "association.subnet-id", "Values": self.subnet_ids},
            ]
        )

        # Consider a subnet to be public if it has a route that targets
        # an internet gateway; private subnets have routes that target NAT gateways
        for route_table in route_tables:
            if any(route.nat_gateway_id for route in route_table.routes):
                return "DISABLED"
        return "ENABLED"

    def get_task_logs(self, task_arn, container_name, limit=10):
        task = self.ecs.describe_tasks(cluster=self.cluster_name, tasks=[task_arn]).get("tasks")[0]

        task_definition_arn = task.get("taskDefinitionArn")
        task_definition = self.ecs.describe_task_definition(taskDefinition=task_definition_arn).get(
            "taskDefinition"
        )

        matching_container_definitions = [
            container_definition
            for container_definition in task_definition.get("containerDefinitions", [])
            if container_definition["name"] == container_name
        ]
        if not matching_container_definitions:
            raise Exception(f"Could not find container with name {container_name}")

        container_definition = matching_container_definitions[0]

        log_stream_prefix = (
            container_definition.get("logConfiguration").get("options").get("awslogs-stream-prefix")
        )
        container_name = container_definition.get("name")
        task_id = task_arn.split("/")[-1]

        log_stream = f"{log_stream_prefix}/{container_name}/{task_id}"

        events = self.logs.get_log_events(
            logGroupName=self.log_group, logStreamName=log_stream, limit=limit
        ).get("events")

        return [event.get("message") for event in events]



ANOMALY_DETECTION_INFERENCE_MUTATION = """
mutation AnomalyDetectionInferenceMutation($modelVersion: String!, $params: GenericScalar!) {
  anomalyDetectionInference(modelVersion: $modelVersion, params: $params) {
    __typename
    ... on AnomalyDetectionSuccess {
      response
    }
    ... on AnomalyDetectionFailure {
      message
    }
    ... on PythonError {
      message
      stack
    }
  }
}
"""



from dagster._core.storage.tags import HIDDEN_TAG_PREFIX, SYSTEM_TAG_PREFIX

CLOUD_SYSTEM_TAG_PREFIX = "dagster-cloud/"

PEX_METADATA_TAG = f"{HIDDEN_TAG_PREFIX}pex_metadata"
PEX_TAG_TAG = f"{SYSTEM_TAG_PREFIX}pex_tag"
IGNORE_ALERTS_TAG = f"{CLOUD_SYSTEM_TAG_PREFIX}ignore-alerts"


def get_policy_names_from_tag_value(policies_str):
    """From a comma-delineated string (whitespace allowed), retrieve alert policy names."""
    return [policy_name.strip() for policy_name in policies_str.split(",")]



from dagster_k8s import K8sRunLauncher


class CloudK8sRunLauncher(K8sRunLauncher):
    # Avoid call to `count_resume_run_attempts`, since resuming runs is not currently
    # supported in cloud and the method makes repeated event log calls.
    @property
    def supports_run_worker_crash_recovery(self):
        return False



import logging
import os
import subprocess
import sys
import threading
from collections import defaultdict
from typing import Any, Collection, Dict, List, Mapping, NamedTuple, Optional, Set, Tuple, Union

import dagster._seven as seven
from dagster import (
    BoolSource,
    Field,
    IntSource,
    _check as check,
)
from dagster._core.errors import DagsterUserCodeUnreachableError
from dagster._grpc.client import DagsterGrpcClient, client_heartbeat_thread
from dagster._serdes import ConfigurableClass, ConfigurableClassData
from dagster._serdes.ipc import open_ipc_subprocess
from dagster._utils import find_free_port, safe_tempfile_path_unmanaged
from dagster._utils.merger import merge_dicts
from dagster_cloud_cli.core.workspace import CodeDeploymentMetadata
from typing_extensions import Self

from dagster_cloud.api.dagster_cloud_api import (
    UserCodeDeploymentType,
)
from dagster_cloud.execution.cloud_run_launcher.process import CloudProcessRunLauncher
from dagster_cloud.execution.monitoring import CloudContainerResourceLimits
from dagster_cloud.pex.grpc import MultiPexGrpcClient

from .user_code_launcher import (
    DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
    SHARED_USER_CODE_LAUNCHER_CONFIG,
    DagsterCloudGrpcServer,
    DagsterCloudUserCodeLauncher,
    ServerEndpoint,
    UserCodeLauncherEntry,
)

CLEANUP_ZOMBIE_PROCESSES_INTERVAL = 5


class ProcessUserCodeEntry(
    NamedTuple(
        "_ProcessUserCodeEntry",
        [
            ("grpc_server_process", subprocess.Popen),
            ("grpc_client", DagsterGrpcClient),
            ("heartbeat_shutdown_event", threading.Event),
            ("heartbeat_thread", threading.Thread),
        ],
    )
):
    def __new__(
        cls,
        grpc_server_process: subprocess.Popen,
        grpc_client: DagsterGrpcClient,
        heartbeat_shutdown_event: threading.Event,
        heartbeat_thread: threading.Thread,
    ):
        return super(ProcessUserCodeEntry, cls).__new__(
            cls,
            check.inst_param(grpc_server_process, "grpc_server_process", subprocess.Popen),
            check.inst_param(grpc_client, "grpc_client", DagsterGrpcClient),
            check.inst_param(heartbeat_shutdown_event, "heartbeat_shutdown_event", threading.Event),
            check.inst_param(heartbeat_thread, "heartbeat_thread", threading.Thread),
        )


class MultipexUserCodeEntry(
    NamedTuple(
        "_MultipexUserCodeEntry",
        [
            ("grpc_server_process", subprocess.Popen),
            ("grpc_client", MultiPexGrpcClient),
        ],
    )
):
    def __new__(
        cls,
        grpc_server_process: subprocess.Popen,
        grpc_client: MultiPexGrpcClient,
    ):
        return super(MultipexUserCodeEntry, cls).__new__(
            cls,
            check.inst_param(grpc_server_process, "grpc_server_process", subprocess.Popen),
            check.inst_param(grpc_client, "grpc_client", MultiPexGrpcClient),
        )


class ProcessUserCodeLauncher(DagsterCloudUserCodeLauncher, ConfigurableClass):
    def __init__(
        self,
        inst_data: Optional[ConfigurableClassData] = None,
        wait_for_processes: bool = False,
        **kwargs,
    ):
        self._inst_data = inst_data
        self._logger = logging.getLogger("dagster_cloud")

        # map from pid to server being spun up
        # (including old servers in the process of being shut down)
        self._process_entries: Dict[int, Union[ProcessUserCodeEntry, MultipexUserCodeEntry]] = {}

        # map from locationname to the pid(s) for that location-metadata combination.
        # Generally there should be only one pid per location unless an exception was raised partway
        # through an update
        self._active_pids: Dict[Tuple[str, str], Set[int]] = defaultdict(set)

        self._active_multipex_pids: Dict[Tuple[str, str], Set[int]] = defaultdict(set)

        self._heartbeat_ttl = 60
        self._wait_for_processes = wait_for_processes

        self._cleanup_zombies_shutdown_event = threading.Event()
        self._cleanup_zombies_thread = None

        self._run_launcher: Optional[CloudProcessRunLauncher] = None

        super(ProcessUserCodeLauncher, self).__init__(**kwargs)

    @property
    def requires_images(self) -> bool:
        return False

    def get_code_server_resource_limits(
        self, deployment_name: str, location_name: str
    ) -> CloudContainerResourceLimits:
        import psutil

        return {
            "process": {
                "cpu_limit": str(os.cpu_count()),
                "memory_limit": str(psutil.virtual_memory().total),
            },
        }

    def start(self, run_reconcile_thread=True, run_metrics_thread=True):
        super().start(
            run_reconcile_thread=run_reconcile_thread, run_metrics_thread=run_metrics_thread
        )
        # TODO Identify if zombie processes are an issue on Windows and what
        # the proper way to clean them up is
        if sys.platform != "win32":
            self._cleanup_zombies_thread = threading.Thread(
                target=self._cleanup_zombie_processes,
                args=(self._cleanup_zombies_shutdown_event,),
                name="cleanup-zombie-processes",
                daemon=True,
            )
            self._cleanup_zombies_thread.start()

    def _cleanup_zombie_processes(self, shutdown_event):
        while True:
            shutdown_event.wait(CLEANUP_ZOMBIE_PROCESSES_INTERVAL)
            if shutdown_event.is_set():
                break

            # Clean up any child processes that have finished since last check
            while True:
                # This may need to be different on Windows because process groups are
                # handled differently.
                try:
                    # https://docs.python.org/3/library/os.html#os.waitpid
                    # If pid is -1, the request pertains to any child of the current process.
                    pid, _exit_code = os.waitpid(-1, os.WNOHANG)
                except ChildProcessError:
                    # Raised when there are no child processes
                    break

                if pid == 0:
                    break

    @property
    def inst_data(self) -> Optional[ConfigurableClassData]:
        return self._inst_data

    @classmethod
    def config_type(cls) -> Dict:
        return merge_dicts(
            {
                "server_process_startup_timeout": Field(
                    IntSource,
                    is_required=False,
                    default_value=DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
                    description=(
                        "Timeout when waiting for a code server to be ready after it is created"
                    ),
                ),
                "wait_for_processes": Field(
                    BoolSource,
                    is_required=False,
                    default_value=False,
                    description=(
                        "When cleaning up the agent, wait for any subprocesses to "
                        "finish before shutting down. Generally only needed in tests/automation."
                    ),
                ),
                "code_server_metrics": Field(
                    {"enabled": Field(bool, is_required=False, default_value=False)},
                    is_required=False,
                ),
                "agent_metrics": Field(
                    {"enabled": Field(bool, is_required=False, default_value=False)},
                    is_required=False,
                ),
            },
            SHARED_USER_CODE_LAUNCHER_CONFIG,
        )

    @property
    def user_code_deployment_type(self) -> UserCodeDeploymentType:
        return UserCodeDeploymentType.PROCESS

    @classmethod
    def from_config_value(
        cls, inst_data: ConfigurableClassData, config_value: Mapping[str, Any]
    ) -> Self:
        return cls(inst_data=inst_data, **config_value)

    def _start_new_server_spinup(
        self,
        deployment_name: str,
        location_name: str,
        metadata: CodeDeploymentMetadata,
    ) -> DagsterCloudGrpcServer:
        key = (deployment_name, location_name)

        client: Union[MultiPexGrpcClient, DagsterGrpcClient]

        port: Optional[int] = None
        socket: Optional[str] = None

        if seven.IS_WINDOWS:
            port = find_free_port()
            socket = None
        else:
            port = None
            socket = safe_tempfile_path_unmanaged()

        if metadata.pex_metadata:
            multipex_process = open_ipc_subprocess(
                metadata.get_multipex_server_command(
                    port,
                    socket,
                    metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled,
                )
            )

            pid = multipex_process.pid

            client = MultiPexGrpcClient(port=port, socket=socket)

            self._process_entries[pid] = MultipexUserCodeEntry(
                multipex_process,
                client,
            )

            self._active_multipex_pids[key].add(pid)
        else:
            subprocess_args = metadata.get_grpc_server_command(
                metrics_enabled=self._instance.user_code_launcher.code_server_metrics_enabled
            ) + [
                "--heartbeat",
                "--heartbeat-timeout",
                str(self._heartbeat_ttl),
            ]

            additional_env = metadata.get_grpc_server_env(
                port=port,
                location_name=location_name,
                instance_ref=self._instance.ref_for_deployment(deployment_name),
                socket=socket,
            )

            server_process = open_ipc_subprocess(
                subprocess_args,
                env={
                    **os.environ,
                    **{key: str(value) for key, value in metadata.cloud_context_env.items()},
                    **additional_env,
                },
            )
            client = DagsterGrpcClient(
                port=port,
                socket=socket,
                host="localhost",
                use_ssl=False,
            )

            heartbeat_shutdown_event = threading.Event()
            heartbeat_thread = threading.Thread(
                target=client_heartbeat_thread,
                args=(client, heartbeat_shutdown_event),
                daemon=True,
            )
            heartbeat_thread.start()

            pid = server_process.pid

            self._process_entries[server_process.pid] = ProcessUserCodeEntry(
                server_process,
                client,
                heartbeat_shutdown_event,
                heartbeat_thread,
            )

            self._active_pids[key].add(pid)

        endpoint = ServerEndpoint(
            host="localhost",
            port=port,
            socket=socket,
        )

        return DagsterCloudGrpcServer(pid, endpoint, metadata)

    def _wait_for_new_server_ready(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
        server_handle: int,
        server_endpoint: ServerEndpoint,
    ) -> None:
        self._wait_for_dagster_server_process(
            client=server_endpoint.create_client(),
            timeout=self._server_process_startup_timeout,
        )

    def _get_standalone_dagster_server_handles_for_location(
        self, deployment_name: str, location_name: str
    ) -> Collection[int]:
        return self._active_pids.get((deployment_name, location_name), set()).copy()

    def _get_multipex_server_handles_for_location(
        self, deployment_name: str, location_name: str
    ) -> Collection[int]:
        return self._active_multipex_pids.get((deployment_name, location_name), set()).copy()

    def _remove_server_handle(self, server_handle: int) -> None:
        pid = server_handle
        process_entry = self._process_entries[pid]
        self._remove_pid(pid)
        if self._wait_for_processes:
            if isinstance(process_entry, ProcessUserCodeEntry):
                try:
                    process_entry.grpc_client.shutdown_server()
                except DagsterUserCodeUnreachableError:
                    # Server already shutdown
                    pass
            else:
                process_entry.grpc_server_process.terminate()

            if process_entry.grpc_server_process.poll() is None:
                process_entry.grpc_server_process.communicate(timeout=30)

    def _remove_pid(self, pid):
        if pid in self._process_entries:
            process_entry = self._process_entries[pid]
            if isinstance(process_entry, ProcessUserCodeEntry):
                # Rely on heartbeat failure to eventually kill the process
                process_entry.heartbeat_shutdown_event.set()
                process_entry.heartbeat_thread.join()
            else:
                # multi-pex server processes don't yet have heartbeats, so just terminate
                # the multipex server process directly.
                process_entry.grpc_server_process.terminate()

            del self._process_entries[pid]

        for pids in self._active_pids.values():
            if pid in pids:
                pids.remove(pid)

        for pids in self._active_multipex_pids.values():
            if pid in pids:
                pids.remove(pid)

    def run_launcher(self) -> CloudProcessRunLauncher:
        if not self._run_launcher:
            self._run_launcher = CloudProcessRunLauncher()
            self._run_launcher.register_instance(self._instance)

        return self._run_launcher

    def _list_server_handles(self) -> List[int]:
        return list(self._process_entries.keys())

    def get_agent_id_for_server(self, handle: int) -> Optional[str]:
        return self._instance.instance_uuid

    def __exit__(self, exception_type, exception_value, traceback):
        super().__exit__(exception_value, exception_value, traceback)

        if self._cleanup_zombies_thread:
            self._cleanup_zombies_shutdown_event.set()
            self._cleanup_zombies_thread.join()
